# SPDX-FileCopyrightText: 2021 GNOME Foundation
# SPDX-License-Identifier: Apache-2.0 OR GPL-3.0-or-later

import markdown
import os
import re
import subprocess
import sys

from markupsafe import Markup
from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters import HtmlFormatter
from typogrify.filters import typogrify

from . import gir, log, mdext


# The beginning of a gtk-doc code block:
#
# |[ (optional language identifier)
#
CODEBLOCK_START_RE = re.compile(
    r'''
    ^
    \s*
    \|\[
    \s*
    (?P<language>\<\!-- \s* language="\w+" \s* --\>)?
    \s*
    $
    ''',
    re.UNICODE | re.VERBOSE)

# The optional language identifier for a gtk-doc code block:
#
# <!-- language="..." -->
#
LANGUAGE_RE = re.compile(
    r'''
    ^
    \s*
    <!--
    \s*
    language="(?P<language>\w+)"
    \s*
    -->
    \s*
    $
    ''',
    re.UNICODE | re.VERBOSE)

# The ending of a gtk-doc code block:
#
# ]|
#
CODEBLOCK_END_RE = re.compile(
    r'''
    ^
    \s*
    \]\|
    \s*
    $
    ''',
    re.UNICODE | re.VERBOSE)

LINK_RE = re.compile(
    r'''
    (?P<text>\[ [\w\s,\-_:]+ \])?
    \[
    (`)?
    (?P<fragment>[\w]+)
    @
    (?P<endpoint>[\w\-_:\.]+)
    (?P<anchor>\#[\w\-_]+)?
    (`)?
    \]
    ''',
    re.VERBOSE)

TYPE_RE = re.compile(
    r'''
    (?P<ns>[\w]+\.)?            # namespace (optional)
    (?P<name>[\w]+)             # type name
    ''',
    re.VERBOSE)

PROPERTY_RE = re.compile(
    r'''
    (?P<ns>[\w]+\.)?            # namespace (optional)
    (?P<name>[\w]+)             # type name
    :{1}                        # delimiter
    (?P<property>[\w-]*\w)      # property name
    ''',
    re.VERBOSE)

SIGNAL_RE = re.compile(
    r'''
    (?P<ns>[\w]+\.)?            # namespace (optional)
    (?P<name>[\w]+)             # type name
    :{2}                        # delimiter
    (?P<signal>[\w-]*\w)        # signal name
    ''',
    re.VERBOSE)

METHOD_RE = re.compile(
    r'''
    (?P<ns>[\w]+\.)?            # namespace (optional)
    (?P<name>[\w]+)             # type name
    \.                          # delimiter
    (?P<method>[\w_]*\w)        # method name
    ''',
    re.VERBOSE)

LANGUAGE_MAP = {
    'c': 'c',
    'css': 'css',
    'plain': 'plain',
    'xml': 'xml',
    'javascript': 'javascript',
}

MD_EXTENSIONS = [
    # Standard extensions
    'codehilite',
    'def_list',
    'fenced_code',
    'meta',
    'tables',
    'toc',

    # Local extensions
    mdext.GtkDocExtension(),
    mdext.AdmonitionExtension(),
]

MD_EXTENSIONS_CONF = {
    'codehilite': {'guess_lang': False},
    'toc': {'permalink_class': 'md-anchor', 'permalink': ''},
}

EN_STOPWORDS = set("""
a  and  are  as  at
be  but  by
for
if  in  into  is  it
near  no  not
of  on  or
such
that  the  their  then  there  these  they  this  to
was  will  with
""".split())


def process_language(lang):
    if lang is None:
        return "plain"

    res = LANGUAGE_RE.match(lang)
    if res:
        language = res.group("language") or "plain"
    else:
        language = "plain"

    return LANGUAGE_MAP[language.lower()]


def parse_error(msg, line=None, start=0, end=0, fragment=None, rest=None):
    if line is not None:
        res = [msg]
        res.append(line)
        err_line = ['^'.rjust(start + 1, ' ')]
        err_line += [''.join(['~' for x in range(end - start - 1)])]
        res.append("".join(err_line))
        return "\n".join(res)
    elif fragment is not None:
        return f"{msg}: [{fragment}@{rest}]"
    else:
        return f"{msg}: [{rest}]"


class LinkParseError(Exception):
    def __init__(self, line=None, start=0, end=0, fragment=None, rest=None, message="Unable to parse link"):
        self.line = line
        self.start = start
        self.end = end
        self.fragment = fragment
        self.rest = rest
        self.message = message

    def __str__(self):
        return parse_error(self.message, self.line, self.start, self.end, self.fragment, self.rest)


class LinkGenerator:
    def __init__(self, **kwargs):
        self._line = kwargs.get('line')
        self._start = kwargs.get('start', 0)
        self._end = kwargs.get('end', 0)
        self._namespace = kwargs.get('namespace')
        self._fragment = kwargs.get('fragment', '')
        self._endpoint = kwargs.get('endpoint', '')
        self._anchor = kwargs.get('anchor')
        self._no_link = kwargs.get('no_link', False)
        self._alt_text = kwargs.get('text')
        self._do_raise = kwargs.get('do_raise', False)
        self._enum_member_name = None
        self._vfunc_name = None

        assert self._namespace is not None

        self._repository = self._namespace.repository
        self._valid_namespaces = [n for n in self._repository.includes]
        self._external = False

        if self._anchor is not None and self._anchor.startswith('#'):
            self._anchor = self._anchor[1:]

        fragment_parsers = {
            "alias": self._parse_type,
            "callback": self._parse_type,
            "class": self._parse_type,
            "const": self._parse_type,
            "ctor": self._parse_method,
            "enum": self._parse_enum_type,
            "error": self._parse_enum_type,
            "flags": self._parse_enum_type,
            "func": self._parse_func,
            "id": self._parse_id,
            "iface": self._parse_type,
            "method": self._parse_method,
            "property": self._parse_property,
            "signal": self._parse_signal,
            "struct": self._parse_type,
            "type": self._parse_type,
            "vfunc": self._parse_method,
        }

        parser_method = fragment_parsers.get(self._fragment)
        if parser_method is not None:
            try:
                parser_method(self._fragment, self._endpoint)
            except LinkParseError as err:
                if self._do_raise:
                    raise
                else:
                    log.warning(str(err))
                    self._fragment = None
        else:
            if self._do_raise:
                raise LinkParseError(self._line, self._start, self._end,
                                     self._fragment, self._endpoint,
                                     f"Unknown fragment {self._fragment}")
            else:
                log.warning(parse_error(f"Unknown fragment {self._fragment}",
                                        self._line, self._start, self._end,
                                        self._fragment, self._endpoint))
                self._fragment = None

    def _parse_id(self, fragment, endpoint):
        symbol = self._repository.find_symbol(endpoint)
        if symbol is None:
            raise LinkParseError(self._line, self._start, self._end,
                                 fragment, endpoint,
                                 f"Unable to find symbol {self._endpoint}")
        (ns, t) = symbol
        if isinstance(t, gir.Class) or \
           isinstance(t, gir.Interface) or \
           isinstance(t, gir.Record):
            self._external = ns is not self._namespace
            self._ns = ns.name
            self._fragment = 'method'
            self._symbol_name = f"{endpoint}()"
            self._name = t.name
            self._method_name = endpoint.replace(ns.symbol_prefix[0] + '_', '')
            self._method_name = self._method_name.replace(t.symbol_prefix + '_', '')
        elif isinstance(t, gir.Function):
            self._external = ns is not self._namespace
            self._ns = ns.name
            self._fragment = 'func'
            self._symbol_name = f"{endpoint}()"
            self._name = None
            self._func_name = endpoint.replace(ns.symbol_prefix[0] + '_', '')
        else:
            raise LinkParseError(self._line, self._start, self._end,
                                 fragment, endpoint,
                                 f"Unsupported symbol {endpoint}")

    def _parse_type(self, fragment, endpoint):
        res = TYPE_RE.match(endpoint)
        if res:
            ns = res.group('ns')
            name = res.group('name')
            rest = endpoint
            len_ns = len(ns) if ns else 0
            len_name = len(name) if name else 0
            rest = endpoint[len_ns + len_name:]
            if ns is not None and name is None:
                name = ns
                ns = None
            if ns is not None:
                ns = ns[:len(ns) - 1]   # Drop the trailing dot
            else:
                ns = self._namespace.name
                # Accept FooBar in place of Foo.Bar
                if name.startswith(tuple(self._namespace.identifier_prefix)):
                    for prefix in self._namespace.identifier_prefix:
                        name = name.replace(prefix, '')
        else:
            raise LinkParseError(self._line, self._start, self._end,
                                 fragment, endpoint,
                                 "Invalid type link")
        if fragment in ['alias', 'callback', 'class', 'const', 'iface', 'struct', 'type'] and rest:
            raise LinkParseError(self._line, self._start, self._end,
                                 fragment, endpoint,
                                 f"Unknown component in {fragment} link for {ns}.{name}: {rest}")
        if ns == self._namespace.name:
            namespace = self._namespace
            self._external = False
            self._ns = ns
        else:
            repository = self._namespace.repository
            namespace = repository.find_included_namespace(ns)
            if namespace is not None:
                self._external = True
                self._ns = namespace.name
            else:
                raise LinkParseError(self._line, self._start, self._end,
                                     fragment, endpoint,
                                     f"Unknown namespace {ns}")
        t = namespace.find_real_type(name)
        if t is not None and t.base_ctype is not None:
            # We determine the fragment here, in case `type` was used,
            # or for validating the fragment passed, to avoid creating
            # invalid links
            if isinstance(t, gir.Alias):
                type_fragment = 'alias'
            elif isinstance(t, gir.Callback):
                type_fragment = 'callback'
            elif isinstance(t, gir.Class):
                type_fragment = 'class'
            elif isinstance(t, gir.Constant):
                type_fragment = 'const'
            elif isinstance(t, gir.Enumeration):
                if isinstance(t, gir.BitField):
                    type_fragment = 'flags'
                elif isinstance(t, gir.ErrorDomain):
                    type_fragment = 'error'
                else:
                    type_fragment = 'enum'
            elif isinstance(t, gir.Interface):
                type_fragment = 'iface'
            elif isinstance(t, gir.Record) or isinstance(t, gir.Union):
                type_fragment = 'struct'
            else:
                raise LinkParseError(self._line, self._start, self._end,
                                     fragment, endpoint,
                                     f"Invalid type {t} for '{ns}.{name}'")
            if fragment != 'type' and fragment != type_fragment:
                raise LinkParseError(self._line, self._start, self._end,
                                     fragment, endpoint,
                                     f"Invalid fragment for '{ns}.{name}': it should be {type_fragment}")
            self._fragment = type_fragment
            self._name = name
            self._type = t.base_ctype
        else:
            raise LinkParseError(self._line, self._start, self._end,
                                 fragment, endpoint,
                                 f"Unable to find type '{ns}.{name}'")

    def _parse_enum_type(self, fragment, endpoint):
        res = TYPE_RE.match(endpoint)
        if res:
            ns = res.group('ns')
            name = res.group('name')
            rest = endpoint
            len_ns = len(ns) if ns else 0
            len_name = len(name) if name else 0
            rest = endpoint[len_ns + len_name:]
            if ns is not None and name is None:
                name = ns
                ns = None
            if ns is not None:
                ns = ns[:len(ns) - 1]   # Drop the trailing dot
            else:
                ns = self._namespace.name
                # Accept FooBar in place of Foo.Bar
                if name.startswith(tuple(self._namespace.identifier_prefix)):
                    for prefix in self._namespace.identifier_prefix:
                        name = name.replace(prefix, '')
        else:
            raise LinkParseError(self._line, self._start, self._end,
                                 fragment, endpoint,
                                 "Invalid type link")
        if ns == self._namespace.name:
            namespace = self._namespace
            self._external = False
            self._ns = ns
        else:
            repository = self._namespace.repository
            namespace = repository.find_included_namespace(ns)
            if namespace is not None:
                self._external = True
                self._ns = namespace.name
            else:
                raise LinkParseError(self._line, self._start, self._end,
                                     fragment, endpoint,
                                     f"Unknown namespace {ns}")
        t = namespace.find_real_type(name)
        if t is not None and t.base_ctype is not None:
            if isinstance(t, gir.Enumeration):
                if isinstance(t, gir.BitField):
                    type_fragment = 'flags'
                elif isinstance(t, gir.ErrorDomain):
                    type_fragment = 'error'
                else:
                    type_fragment = 'enum'
            else:
                raise LinkParseError(self._line, self._start, self._end,
                                     fragment, endpoint,
                                     f"Invalid type {t} for '{ns}.{name}'")
            if fragment != type_fragment:
                raise LinkParseError(self._line, self._start, self._end,
                                     fragment, endpoint,
                                     f"Invalid fragment for '{ns}.{name}': it should be {type_fragment}")
            if rest:
                if not rest.startswith('.'):
                    raise LinkParseError(self._line, self._start, self._end,
                                         fragment, endpoint,
                                         f"Invalid member for enumeration {ns}.{name}")
                e = rest[1:len(rest)]
                uc_member = e.upper().replace('-', '_')
                found = False
                for member in t:
                    if member.name.upper() == uc_member:
                        self._anchor = member.nick
                        self._enum_member_name = member.identifier
                        found = True
                        break
                if not found:
                    raise LinkParseError(self._line, self._start, self._end,
                                         fragment, endpoint,
                                         f"Invalid member {e} for enumeration {ns}.{name}")
            self._fragment = type_fragment
            self._name = name
            self._type = t.base_ctype
        else:
            raise LinkParseError(self._line, self._start, self._end,
                                 fragment, endpoint,
                                 f"Unable to find type '{ns}.{name}'")

    def _parse_property(self, fragment, endpoint):
        res = PROPERTY_RE.match(endpoint)
        if res:
            ns = res.group('ns')
            name = res.group('name')
            pname = res.group('property')
            if ns is not None:
                ns = ns[:len(ns) - 1]   # Drop the trailing dot
            else:
                ns = self._namespace.name
                # Accept FooBar in place of Foo.Bar
                if name.startswith(tuple(self._namespace.identifier_prefix)):
                    for prefix in self._namespace.identifier_prefix:
                        name = name.replace(prefix, '')
            # Canonicalize the property name
            pname = pname.replace('_', '-')
        else:
            raise LinkParseError(self._line, self._start, self._end,
                                 fragment, endpoint,
                                 "Invalid property link")
        if ns == self._namespace.name:
            namespace = self._namespace
            self._external = False
            self._ns = ns
        else:
            repository = self._namespace.repository
            namespace = repository.find_included_namespace(ns)
            if namespace is not None:
                self._external = True
                self._ns = ns
            else:
                self._fragment = None
                return
        t = namespace.find_real_type(name)
        if t is not None and t.base_ctype is not None:
            self._type = t.base_ctype
            self._name = name
        else:
            raise LinkParseError(self._line, self._start, self._end,
                                 fragment, endpoint,
                                 f"Unable to find type '{ns}.{name}'")
        if (isinstance(t, gir.Class) or isinstance(t, gir.Interface)) and pname in t.properties:
            self._property_name = pname
        else:
            raise LinkParseError(self._line, self._start, self._end,
                                 fragment, endpoint,
                                 f"Invalid property '{pname}' for type '{ns}.{name}'")

    def _parse_signal(self, fragment, endpoint):
        res = SIGNAL_RE.match(endpoint)
        if res:
            ns = res.group('ns')
            name = res.group('name')
            sname = res.group('signal')
            if ns is not None:
                ns = ns[:len(ns) - 1]   # Drop the trailing dot
            else:
                ns = self._namespace.name
                # Accept FooBar in place of Foo.Bar
                if name.startswith(tuple(self._namespace.identifier_prefix)):
                    for prefix in self._namespace.identifier_prefix:
                        name = name.replace(prefix, '')
            # Canonicalize the signal name
            sname = sname.replace('_', '-')
        else:
            raise LinkParseError(self._line, self._start, self._end,
                                 fragment, endpoint,
                                 "Invalid signal link")
        if ns == self._namespace.name:
            namespace = self._namespace
            self._external = False
            self._ns = ns
        else:
            repository = self._namespace.repository
            namespace = repository.find_included_namespace(ns)
            if namespace is not None:
                self._external = True
                self._ns = namespace.name
            else:
                self._fragment = None
                return
        t = namespace.find_real_type(name)
        if t is not None and t.base_ctype is not None:
            self._type = t.base_ctype
            self._name = name
        else:
            raise LinkParseError(self._line, self._start, self._end,
                                 fragment, endpoint,
                                 f"Unable to find type '{ns}.{name}'")
        if (isinstance(t, gir.Class) or isinstance(t, gir.Interface)) and sname in t.signals:
            self._signal_name = sname
        else:
            raise LinkParseError(self._line, self._start, self._end,
                                 fragment, endpoint,
                                 f"Invalid signal name '{sname}' for type '{ns}.{name}'")

    def _parse_method(self, fragment, endpoint):
        res = METHOD_RE.match(endpoint)
        if res:
            ns = res.group('ns')
            name = res.group('name')
            method = res.group('method')
            if ns is not None:
                ns = ns[:len(ns) - 1]   # Drop the trailing dot
            else:
                ns = self._namespace.name
                # Accept FooBar in place of Foo.Bar
                if name.startswith(tuple(self._namespace.identifier_prefix)):
                    for prefix in self._namespace.identifier_prefix:
                        name = name.replace(prefix, '')
        else:
            raise LinkParseError(self._line, self._start, self._end,
                                 fragment, endpoint,
                                 "Invalid method link")
        if ns == self._namespace.name:
            namespace = self._namespace
            self._external = False
            self._ns = ns
        else:
            repository = self._namespace.repository
            namespace = repository.find_included_namespace(ns)
            if namespace is not None:
                self._ns = namespace.name
                self._external = True
            else:
                self._fragment = None
                return
        t = namespace.find_real_type(name)
        if t is not None and t.base_ctype is not None:
            self._type = t.base_ctype
            self._method_name = method
            # method@Foo.BarClass.add_name -> class_method.Bar.add_name.html
            if isinstance(t, gir.Record) and t.struct_for is not None:
                self._name = t.struct_for
                self._fragment = "class_method"
            elif fragment == "vfunc" and t.type_struct is not None:
                self._name = name
                self._type = t.type_struct
            else:
                self._name = name
        else:
            raise LinkParseError(self._line, self._start, self._end,
                                 fragment, endpoint,
                                 f"Unable to find type '{ns}.{name}'")
        if fragment == "ctor":
            methods = getattr(t, "constructors", [])
        elif fragment in ["method", "class_method"]:
            methods = getattr(t, "methods", [])
        elif fragment == "vfunc":
            methods = getattr(t, "virtual_methods", [])
        else:
            methods = []
        for m in methods:
            if m.name == method:
                if fragment == "vfunc":
                    self._vfunc_name = m.name
                else:
                    self._symbol_name = f"{m.identifier}()"
                return
        raise LinkParseError(self._line, self._start, self._end,
                             fragment, endpoint,
                             f"Unable to find method '{ns}.{name}.{method}'")

    def _parse_func(self, fragment, endpoint):
        tokens = endpoint.split('.')
        # Case 1: [func@init] => gtk_init()
        if len(tokens) == 1:
            ns = self._namespace.name
            name = None
            func_name = tokens[0]
        # Case 2: [func@Gtk.Foo.bar] => gtk_foo_bar()
        elif len(tokens) == 3:
            ns = tokens[0]
            name = tokens[1]
            func_name = tokens[2]
        # Case 3: either [func@Gtk.init] or [func@Foo.bar]
        elif len(tokens) == 2:
            if tokens[0] == self._namespace.name:
                ns = tokens[0]
                name = None
                func_name = tokens[1]
            elif tokens[0] in self._valid_namespaces:
                ns = tokens[0]
                name = None
                func_name = tokens[1]
            else:
                ns = self._namespace.name
                name = tokens[0]
                func_name = tokens[1]
        else:
            raise LinkParseError(self._line, self._start, self._end,
                                 fragment, endpoint,
                                 "Invalid function link")
        if ns == self._namespace.name:
            namespace = self._namespace
            self._external = False
            self._ns = ns
        else:
            repository = self._namespace.repository
            namespace = repository.find_included_namespace(ns)
            if namespace is not None:
                self._external = True
                self._ns = namespace.name
            else:
                raise LinkParseError(self._line, self._start, self._end,
                                     fragment, endpoint,
                                     f"Namespace {ns} not found")
        if name is None:
            t = namespace.find_function(func_name)
            if t is not None:
                self._name = None
                self._func_name = func_name
                self._symbol_name = f"{t.identifier}()"
            else:
                raise LinkParseError(self._line, self._start, self._end,
                                     fragment, endpoint,
                                     f"Unable to find function '{ns}.{func_name}'")
        else:
            t = namespace.find_real_type(name)
            if t is None:
                raise LinkParseError(self._line, self._start, self._end,
                                     fragment, endpoint,
                                     f"Unable to find type '{ns}.{name}'")
            for func in t.functions:
                if func.name == func_name:
                    self._name = name
                    self._func_name = func.name
                    self._symbol_name = f"{func.identifier}()"
                    return
            raise LinkParseError(self._line, self._start, self._end,
                                 fragment, endpoint,
                                 f"Unable to find function '{ns}.{name}.{func_name}'")

    @property
    def text(self):
        if self._alt_text is not None:
            return self._alt_text[1:len(self._alt_text) - 1]
        elif self._fragment in ['alias', 'callback', 'class', 'const', 'iface', 'struct']:
            return f"<code>{self._type}</code>"
        elif self._fragment in ['enum', 'error', 'flags']:
            if self._enum_member_name:
                return f"<code>{self._enum_member_name}</code>"
            else:
                return f"<code>{self._type}</code>"
        elif self._fragment == 'property':
            return f"<code>{self._type}:{self._property_name}</code>"
        elif self._fragment == 'signal':
            return f"<code>{self._type}::{self._signal_name}</code>"
        elif self._fragment in ['ctor', 'func', 'method', 'class_method']:
            return f"<code>{self._symbol_name}</code>"
        elif self._fragment == 'vfunc':
            return f"<code>{self._ns}.{self._type}.{self._vfunc_name}</code>"
        else:
            return f"{self._endpoint}"

    @property
    def href(self):
        if self._anchor is not None:
            anchor = f"#{self._anchor}"
        else:
            anchor = ""
        if self._fragment in ['alias', 'callback', 'class', 'const', 'enum', 'error', 'flags', 'iface', 'struct']:
            return f"{self._fragment}.{self._name}.html{anchor}"
        elif self._fragment == 'property':
            return f"property.{self._name}.{self._property_name}.html{anchor}"
        elif self._fragment == 'signal':
            return f"signal.{self._name}.{self._signal_name}.html{anchor}"
        elif self._fragment in ['ctor', 'method', 'class_method', 'vfunc']:
            return f"{self._fragment}.{self._name}.{self._method_name}.html{anchor}"
        elif self._fragment == 'func':
            if self._name is not None:
                return f"type_func.{self._name}.{self._func_name}.html{anchor}"
            else:
                return f"func.{self._func_name}.html{anchor}"
        else:
            return None

    def __str__(self):
        text = self.text
        if self._no_link:
            return text
        link = self.href
        if link is None:
            return text
        if self._external:
            data_namespace = f"data-namespace=\"{self._ns}\""
            data_link = f"data-link=\"{link}\""
            href = "href=\"javascript:void(0)\""
            css = "class=\"external\""
            return f"<a {href} {data_namespace} {data_link} {css}>{text}</a>"
        else:
            return f"<a href=\"{link}\">{text}</a>"


def preprocess_docs(text, namespace, summary=False, md=None, extensions=[], plain=False, max_length=20):
    if plain:
        text = text.replace('\n', ' ')
        text = re.sub(r'<[^<]+?>', '', text)
        if max_length > 0:
            words = text.split(' ')
            if len(words) > max_length:
                words = words[:max_length - 1]
                words.append('...')
                text = ' '.join(words)
        return text

    processed_text = []

    code_block_text = []
    code_block_language = None
    inside_code_block = False

    for line in text.split("\n"):
        # If we're in "summary" mode, we bail out at the first empty line
        # after a paragraph
        if summary and line == '' and len(processed_text) > 0:
            break

        res = CODEBLOCK_START_RE.match(line)
        if res:
            code_block_language = process_language(res.group("language"))
            inside_code_block = True
            continue

        res = CODEBLOCK_END_RE.match(line)
        if res and inside_code_block:
            if code_block_language == "plain":
                processed_text += ["```"]
                processed_text.extend(code_block_text)
                processed_text += ["```"]
            else:
                lexer = get_lexer_by_name(code_block_language)
                formatter = HtmlFormatter()
                code_block = highlight("\n".join(code_block_text), lexer, formatter)
                processed_text += [""]
                processed_text.extend(code_block.split("\n"))
                processed_text += [""]

            code_block_language = None
            code_block_text = []
            inside_code_block = False
            continue

        if inside_code_block:
            code_block_text.append(line)
        else:
            new_line = []
            idx = 0
            for m in LINK_RE.finditer(line, idx):
                fragment = m.group('fragment')
                endpoint = m.group('endpoint')
                anchor = m.group('anchor')
                text = m.group('text')
                start = m.start()
                end = m.end()
                link = LinkGenerator(line=line, start=start, end=end,
                                     namespace=namespace,
                                     fragment=fragment, endpoint=endpoint,
                                     anchor=anchor,
                                     no_link=summary, text=text)
                left_pad = line[idx:start]
                replacement = re.sub(LINK_RE, str(link), line[start:end])
                new_line.append(left_pad)
                new_line.append(replacement)
                idx = end

            new_line.append(line[idx:])

            if len(new_line) == 0:
                processed_text.append(line)
            else:
                processed_text.append("".join(new_line))

    if len(processed_text) == 0:
        return ''

    # Capitalize the first character of the first line, but only if it does not
    # start with a link or a gtk-doc marker, to avoid messing up the rest of
    # the string
    first_line = processed_text[0]
    if first_line and first_line[0].isalpha():
        processed_text[0] = ''.join([first_line[0:1].upper(), first_line[1:]])

    # Append a period, if one isn't there already, but not after any code block
    last_line = processed_text[-1]
    if last_line and not last_line.endswith((".", "?", "!", "```")):
        processed_text[-1] = ''.join([last_line, '.'])

    if md is None:
        md_ext = extensions.copy()
        md_ext.extend(MD_EXTENSIONS)
        text = markdown.markdown("\n".join(processed_text),
                                 extensions=md_ext,
                                 extension_configs=MD_EXTENSIONS_CONF)
    else:
        text = md.reset().convert("\n".join(processed_text))

    return Markup(typogrify(text, ignore_tags=['h1', 'h2', 'h3', 'h4']))


def code_highlight(text, language='c'):
    lexer = get_lexer_by_name(language)
    formatter = HtmlFormatter()
    return Markup(highlight(text, lexer, formatter))


def render_dot(dot, output_format="svg"):
    if output_format not in ["svg", "png"]:
        log.error("Invalid output format for render_dot(): {output_format}")
    if output_format == "svg":
        output_arg = "-Tsvg_inline"
    else:
        output_arg = f"-T{output_format}"

    dot_bin = find_program("dot")
    if not dot_bin:
        return None

    args = [dot_bin, output_arg]

    try:
        proc = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        proc.stdin.write(dot.encode("utf-8"))
        output, err = proc.communicate()
        if proc.returncode:
            log.warning(f"Unable to process dot data: {err}")
            return None
        return output.decode("utf-8")
    except Exception as e:
        log.warning(f"Unable to process dot data: {e}")
        return None


found_programs = {}


def find_program(bin_name, path=None, error_if_not_found=False):
    """Finds a program @bin_name inside the given @path, and returns
    its full path if found, or None if the program could not be found.

    The @bin_name will automatically get an extension depending on the
    platform.

    If @error_if_not_found is True, then we'll log an error.
    """
    global found_programs

    if path is None and bin_name in found_programs:
        return found_programs[bin_name]

    if path is None:
        search_paths = os.environ['PATH'].split(os.pathsep)
    else:
        search_paths = path.split(os.pathsep)

    bin_extensions = ['']

    if sys.platform == 'win32':
        pathext = os.environ['PATHEXT'].lower().split(os.pathsep)
        (basename, extension) = os.path.splitext(bin_name)
        if extension.lower() not in pathext:
            bin_extensions = pathext
        search_paths.insert(0, '')

    for ext in bin_extensions:
        executable = bin_name + ext

        for p in search_paths:
            full_path = os.path.join(p, executable)
            if os.path.isfile(full_path):
                # Memoize the result with the default PATH, so we can
                # call this multiple times at no additional cost
                if path is None:
                    found_programs[bin_name] = full_path
                return full_path

    if error_if_not_found:
        log.error(f"Unable to find {bin_name}")

    return None


def default_search_paths():
    if not sys.platform == 'win32':
        xdg_data_dirs = os.environ.get("XDG_DATA_DIRS", "/usr/share:/usr/local/share").split(":")
        xdg_data_home = os.environ.get("XDG_DATA_HOME", os.path.expanduser("~/.local/share"))
    else:
        xdg_data_dirs = None
        xdg_data_home = None

    paths = []
    paths.append(os.getcwd())
    gi_gir_path = os.environ.get('GI_GIR_PATH')
    if gi_gir_path is not None:
        paths.extend(gi_gir_path.split(os.pathsep))
    # Add sys.base_prefix when using MSYS2
    if sys.platform == 'win32' and 'GCC' in sys.version:
        paths.append(os.path.join(sys.base_prefix, 'share', 'gir-1.0'))
    if xdg_data_home is not None:
        paths.append(os.path.join(xdg_data_home, "gir-1.0"))
    if xdg_data_dirs is not None:
        paths.extend([os.path.join(x, "gir-1.0") for x in xdg_data_dirs])
    if sys.platform != 'win32' and '/usr/share/gir-1.0' not in paths:
        paths.append('/usr/share/gir-1.0')

    return paths


def find_extra_content_file(content_dirs, file):
    for p in content_dirs:
        full_path = os.path.join(p, file)
        if os.path.isfile(full_path):
            return full_path

    raise FileNotFoundError(f"Content file {file} not found in any content directory")



# SPDX-FileCopyrightText: 2020 GNOME Foundation
# SPDX-License-Identifier: Apache-2.0 OR GPL-3.0-or-later

import argparse
import json
import os
import sys

from . import config, core, gir, log, utils


HELP_MSG = "Generates the symbol indices for search"

MISSING_DESCRIPTION = "No description available."


def _gen_aliases(config, index, repository, symbols):
    index_symbols = index["symbols"]

    for alias in symbols:
        if config.is_hidden(alias.name):
            log.debug(f"Skipping hidden type {alias.name}")
            continue
        if alias.doc is not None:
            description = alias.doc.content
        else:
            description = MISSING_DESCRIPTION
        if alias.deprecated:
            (deprecated, _) = alias.deprecated_since
        else:
            deprecated = None
        index_symbols.append({
            "type": "alias",
            "name": alias.name,
            "ctype": alias.base_ctype,
            "summary": utils.preprocess_docs(description, repository.namespace, summary=True, plain=True),
            "deprecated": deprecated,
        })


def _gen_bitfields(config, index, repository, symbols):
    index_symbols = index["symbols"]

    for bitfield in symbols:
        if config.is_hidden(bitfield.name):
            log.debug(f"Skipping hidden type {bitfield.name}")
            continue
        if bitfield.doc is not None:
            description = bitfield.doc.content
        else:
            description = MISSING_DESCRIPTION
        if bitfield.deprecated:
            (deprecated, _) = bitfield.deprecated_since
        else:
            deprecated = None
        index_symbols.append({
            "type": "bitfield",
            "name": bitfield.name,
            "ctype": bitfield.base_ctype,
            "summary": utils.preprocess_docs(description, repository.namespace, summary=True, plain=True),
            "deprecated": deprecated,
        })

        for func in bitfield.functions:
            if func.doc is not None:
                func_desc = func.doc.content
            else:
                func_desc = MISSING_DESCRIPTION
            if func.deprecated:
                (func_deprecated, _) = func.deprecated_since
            else:
                func_deprecated = None
            index_symbols.append({
                "type": "type_func",
                "name": func.name,
                "type_name": bitfield.name,
                "ident": func.identifier,
                "summary": utils.preprocess_docs(func_desc, repository.namespace, summary=True, plain=True),
                "deprecated": func_deprecated,
            })


def _gen_callbacks(config, index, repository, symbols):
    index_symbols = index["symbols"]

    for callback in symbols:
        if config.is_hidden(callback.name):
            log.debug(f"Skipping hidden callback {callback.name}")
            continue
        if callback.doc is not None:
            cb_desc = callback.doc.content
        else:
            cb_desc = MISSING_DESCRIPTION
        if callback.deprecated:
            (cb_deprecated, _) = callback.deprecated_since
        else:
            cb_deprecated = None
        index_symbols.append({
            "type": "callback",
            "name": callback.name,
            "ctype": callback.base_ctype,
            "summary": utils.preprocess_docs(cb_desc, repository.namespace, summary=True, plain=True),
            "deprecated": cb_deprecated,
        })


def _gen_classes(config, index, repository, symbols):
    namespace = repository.namespace
    index_symbols = index["symbols"]

    for cls in symbols:
        if config.is_hidden(cls.name):
            log.debug(f"Skipping hidden type {cls.name}")
            continue
        if cls.doc is not None:
            cls_desc = cls.doc.content
        else:
            cls_desc = MISSING_DESCRIPTION
        if cls.deprecated:
            (cls_deprecated, _) = cls.deprecated_since
        else:
            cls_deprecated = None
        index_symbols.append({
            "type": "class",
            "name": cls.name,
            "ctype": cls.base_ctype,
            "summary": utils.preprocess_docs(cls_desc, repository.namespace, summary=True, plain=True),
            "deprecated": cls_deprecated,
        })

        for ctor in cls.constructors:
            if ctor.doc is not None:
                ctor_desc = ctor.doc.content
            else:
                ctor_desc = MISSING_DESCRIPTION
            if ctor.deprecated:
                (ctor_deprecated, _) = ctor.deprecated_since
            else:
                ctor_deprecated = None
            index_symbols.append({
                "type": "ctor",
                "name": ctor.name,
                "type_name": cls.name,
                "ident": ctor.identifier,
                "summary": utils.preprocess_docs(ctor_desc, repository.namespace, summary=True, plain=True),
                "deprecated": ctor_deprecated,
            })

        for method in cls.methods:
            if method.doc is not None:
                method_desc = method.doc.content
            else:
                method_desc = MISSING_DESCRIPTION
            if method.deprecated:
                (method_deprecated, _) = method.deprecated_since
            else:
                method_deprecated = None
            index_symbols.append({
                "type": "method",
                "name": method.name,
                "type_name": cls.name,
                "ident": method.identifier,
                "summary": utils.preprocess_docs(method_desc, repository.namespace, summary=True, plain=True),
                "deprecated": method_deprecated,
            })

        for func in cls.functions:
            if func.doc is not None:
                func_desc = func.doc.content
            else:
                func_desc = MISSING_DESCRIPTION
            if func.deprecated:
                (func_deprecated, _) = func.deprecated_since
            else:
                func_deprecated = None
            index_symbols.append({
                "type": "type_func",
                "name": func.name,
                "type_name": cls.name,
                "ident": func.identifier,
                "summary": utils.preprocess_docs(func_desc, repository.namespace, summary=True, plain=True),
                "deprecated": func_deprecated,
            })

        for prop_name, prop in cls.properties.items():
            if config.is_hidden(cls.name, 'property', prop_name):
                log.debug(f"Skipping hidden property {cls.name}.{prop_name}")
                continue
            if prop.doc is not None:
                prop_desc = prop.doc.content
            else:
                prop_desc = MISSING_DESCRIPTION
            if prop.deprecated:
                (prop_deprecated, _) = prop.deprecated_since
            else:
                prop_deprecated = None
            index_symbols.append({
                "type": "property",
                "name": prop.name,
                "type_name": cls.name,
                "summary": utils.preprocess_docs(prop_desc, repository.namespace, summary=True, plain=True),
                "deprecated": prop_deprecated,
            })

        for signal_name, signal in cls.signals.items():
            if config.is_hidden(cls.name, 'signal', signal_name):
                log.debug(f"Skipping hidden signal {cls.name}.{signal_name}")
                continue
            if signal.doc is not None:
                signal_desc = signal.doc.content
            else:
                signal_desc = MISSING_DESCRIPTION
            if signal.deprecated:
                (signal_deprecated, _) = signal.deprecated_since
            else:
                signal_deprecated = None
            index_symbols.append({
                "type": "signal",
                "name": signal.name,
                "type_name": cls.name,
                "summary": utils.preprocess_docs(signal_desc, repository.namespace, summary=True, plain=True),
                "deprecated": signal_deprecated,
            })

        for vfunc in cls.virtual_methods:
            if vfunc.doc is not None:
                vfunc_desc = vfunc.doc.content
            else:
                vfunc_desc = MISSING_DESCRIPTION
            if vfunc.deprecated:
                (vfunc_deprecated, _) = vfunc.deprecated_since
            else:
                vfunc_deprecated = None
            index_symbols.append({
                "type": "vfunc",
                "name": vfunc.name,
                "type_name": cls.name,
                "summary": utils.preprocess_docs(vfunc_desc, repository.namespace, summary=True, plain=True),
                "deprecated": vfunc_deprecated,
            })

        if cls.type_struct is not None:
            cls_struct = namespace.find_record(cls.type_struct)
            for cls_method in cls_struct.methods:
                if cls_method.doc is not None:
                    cls_method_desc = cls_method.doc.content
                else:
                    cls_method_desc = MISSING_DESCRIPTION
                if cls_method.deprecated:
                    (method_deprecated, _) = cls_method.deprecated_since
                else:
                    method_deprecated = None
                index_symbols.append({
                    "type": "class_method",
                    "name": cls_method.name,
                    "type_name": cls_struct.name,
                    "struct_for": cls_struct.struct_for,
                    "ident": cls_method.identifier,
                    "summary": utils.preprocess_docs(cls_method_desc, repository.namespace, summary=True, plain=True),
                    "deprecated": method_deprecated,
                })


def _gen_constants(config, index, repository, symbols):
    index_symbols = index["symbols"]

    for const in symbols:
        if config.is_hidden(const.name):
            log.debug(f"Skipping hidden const {const.name}")
            continue
        if const.doc is not None:
            const_desc = const.doc.content
        else:
            const_desc = MISSING_DESCRIPTION
        if const.deprecated:
            (const_deprecated, _) = const.deprecated_since
        else:
            const_deprecated = None
        index_symbols.append({
            "type": "constant",
            "name": const.name,
            "ident": const.ctype,
            "summary": utils.preprocess_docs(const_desc, repository.namespace, summary=True, plain=True),
            "deprecated": const_deprecated,
        })


def _gen_domains(config, index, repository, symbols):
    index_symbols = index["symbols"]

    for domain in symbols:
        if config.is_hidden(domain.name):
            log.debug(f"Skipping hidden type {domain.name}")
            continue
        if domain.doc is not None:
            domain_desc = domain.doc.content
        else:
            domain_desc = MISSING_DESCRIPTION
        if domain.deprecated:
            (domain_deprecated, _) = domain.deprecated_since
        else:
            domain_deprecated = None
        index_symbols.append({
            "type": "domain",
            "name": domain.name,
            "ctype": domain.base_ctype,
            "summary": utils.preprocess_docs(domain_desc, repository.namespace, summary=True, plain=True),
            "deprecated": domain_deprecated,
        })

        for func in domain.functions:
            if func.doc is not None:
                func_desc = func.doc.content
            else:
                func_desc = MISSING_DESCRIPTION
            if func.deprecated:
                (func_deprecated, _) = func.deprecated_since
            else:
                func_deprecated = None
            index_symbols.append({
                "type": "type_func",
                "name": func.name,
                "type_name": domain.name,
                "ident": func.identifier,
                "summary": utils.preprocess_docs(func_desc, repository.namespace, summary=True, plain=True),
                "deprecated": func_deprecated,
            })


def _gen_enums(config, index, repository, symbols):
    index_symbols = index["symbols"]

    for enum in symbols:
        if config.is_hidden(enum.name):
            log.debug(f"Skipping hidden type {enum.name}")
            continue
        if enum.doc is not None:
            enum_desc = enum.doc.content
        else:
            enum_desc = MISSING_DESCRIPTION
        if enum.deprecated:
            (enum_deprecated, _) = enum.deprecated_since
        else:
            enum_deprecated = None
        index_symbols.append({
            "type": "enum",
            "name": enum.name,
            "ctype": enum.base_ctype,
            "summary": utils.preprocess_docs(enum_desc, repository.namespace, summary=True, plain=True),
            "deprecated": enum_deprecated,
        })

        for func in enum.functions:
            if func.doc is not None:
                func_desc = func.doc.content
            else:
                func_desc = MISSING_DESCRIPTION
            if func.deprecated:
                (func_deprecated, _) = func.deprecated_since
            else:
                func_deprecated = None
            index_symbols.append({
                "type": "type_func",
                "name": func.name,
                "type_name": enum.name,
                "ident": func.identifier,
                "summary": utils.preprocess_docs(func_desc, repository.namespace, summary=True, plain=True),
                "deprecated": func_deprecated,
            })


def _gen_functions(config, index, repository, symbols):
    index_symbols = index["symbols"]

    for func in symbols:
        if config.is_hidden(func.name):
            log.debug(f"Skipping hidden function {func.name}")
            continue
        if func.doc is not None:
            func_desc = func.doc.content
        else:
            func_desc = MISSING_DESCRIPTION
        if func.deprecated:
            (func_deprecated, _) = func.deprecated_since
        else:
            func_deprecated = None
        index_symbols.append({
            "type": "function",
            "name": func.name,
            "ident": func.identifier,
            "summary": utils.preprocess_docs(func_desc, repository.namespace, summary=True, plain=True),
            "deprecated": func_deprecated,
        })


def _gen_function_macros(config, index, repository, symbols):
    index_symbols = index["symbols"]

    for func in symbols:
        if config.is_hidden(func.name):
            log.debug(f"Skipping hidden macro {func.name}")
            continue
        if func.doc is not None:
            func_desc = func.doc.content
        else:
            func_desc = MISSING_DESCRIPTION
        if func.deprecated:
            (func_deprecated, _) = func.deprecated_since
        else:
            func_deprecated = None
        index_symbols.append({
            "type": "function_macro",
            "name": func.name,
            "ident": func.identifier,
            "summary": utils.preprocess_docs(func_desc, repository.namespace, summary=True, plain=True),
            "deprecated": func_deprecated,
        })


def _gen_interfaces(config, index, repository, symbols):
    namespace = repository.namespace
    index_symbols = index["symbols"]

    for iface in symbols:
        if config.is_hidden(iface.name):
            log.debug(f"Skipping hidden type {iface.name}")
            continue
        if iface.doc is not None:
            iface_desc = iface.doc.content
        else:
            iface_desc = MISSING_DESCRIPTION
        if iface.deprecated:
            (iface_deprecated, _) = iface.deprecated_since
        else:
            iface_deprecated = None
        index_symbols.append({
            "type": "interface",
            "name": iface.name,
            "ctype": iface.base_ctype,
            "summary": utils.preprocess_docs(iface_desc, repository.namespace, summary=True, plain=True),
            "deprecated": iface_deprecated,
        })

        for method in iface.methods:
            if method.doc is not None:
                method_desc = method.doc.content
            else:
                method_desc = MISSING_DESCRIPTION
            if method.deprecated:
                (method_deprecated, _) = method.deprecated_since
            else:
                method_deprecated = None
            index_symbols.append({
                "type": "method",
                "name": method.name,
                "type_name": iface.name,
                "ident": method.identifier,
                "summary": utils.preprocess_docs(method_desc, repository.namespace, summary=True, plain=True),
                "deprecated": method_deprecated,
            })

        for func in iface.functions:
            if func.doc is not None:
                func_desc = func.doc.content
            else:
                func_desc = MISSING_DESCRIPTION
            if func.deprecated:
                (func_deprecated, _) = func.deprecated_since
            else:
                func_deprecated = None
            index_symbols.append({
                "type": "type_func",
                "name": func.name,
                "type_name": iface.name,
                "ident": func.identifier,
                "summary": utils.preprocess_docs(func_desc, repository.namespace, summary=True, plain=True),
                "deprecated": func_deprecated,
            })

        for prop_name, prop in iface.properties.items():
            if config.is_hidden(iface.name, 'property', prop_name):
                log.debug(f"Skipping hidden property {iface.name}.{prop_name}")
                continue
            if prop.doc is not None:
                prop_desc = prop.doc.content
            else:
                prop_desc = MISSING_DESCRIPTION
            if prop.deprecated:
                (prop_deprecated, _) = prop.deprecated_since
            else:
                prop_deprecated = None
            index_symbols.append({
                "type": "property",
                "name": prop.name,
                "type_name": iface.name,
                "summary": utils.preprocess_docs(prop_desc, repository.namespace, summary=True, plain=True),
                "deprecated": prop_deprecated,
            })

        for signal_name, signal in iface.signals.items():
            if config.is_hidden(iface.name, 'signal', signal_name):
                log.debug(f"Skipping hidden signal {iface.name}.{signal_name}")
                continue
            if signal.doc is not None:
                signal_desc = signal.doc.content
            else:
                signal_desc = MISSING_DESCRIPTION
            if signal.deprecated:
                (signal_deprecated, _) = signal.deprecated_since
            else:
                signal_deprecated = None
            index_symbols.append({
                "type": "signal",
                "name": signal.name,
                "type_name": iface.name,
                "summary": utils.preprocess_docs(signal_desc, repository.namespace, summary=True, plain=True),
                "deprecated": signal_deprecated,
            })

        for vfunc in iface.virtual_methods:
            if vfunc.doc is not None:
                vfunc_desc = vfunc.doc.content
            else:
                vfunc_desc = MISSING_DESCRIPTION
            if vfunc.deprecated:
                (vfunc_deprecated, _) = vfunc.deprecated_since
            else:
                vfunc_deprecated = None
            index_symbols.append({
                "type": "vfunc",
                "name": vfunc.name,
                "type_name": iface.name,
                "summary": utils.preprocess_docs(vfunc_desc, repository.namespace, summary=True, plain=True),
                "deprecated": vfunc_deprecated,
            })

        if iface.type_struct is not None:
            iface_struct = namespace.find_record(iface.type_struct)
            for iface_method in iface_struct.methods:
                if iface_method.doc is not None:
                    iface_method_desc = iface_method.doc.content
                else:
                    iface_method_desc = MISSING_DESCRIPTION
                if iface_method.deprecated:
                    (method_deprecated, _) = iface_method.deprecated_since
                else:
                    method_deprecated = None
                index_symbols.append({
                    "type": "class_method",
                    "name": iface_method.name,
                    "type_name": iface_struct.name,
                    "struct_for": iface_struct.struct_for,
                    "ident": iface_method.identifier,
                    "summary": utils.preprocess_docs(iface_method_desc, repository.namespace, summary=True, plain=True),
                    "deprecated": method_deprecated,
                })


def _gen_records(config, index, repository, symbols):
    index_symbols = index["symbols"]

    for record in symbols:
        if config.is_hidden(record.name):
            log.debug(f"Skipping hidden type {record.name}")
            continue
        if record.doc is not None:
            desc = record.doc.content
        else:
            desc = MISSING_DESCRIPTION
        if record.deprecated:
            (deprecated, _) = record.deprecated_since
        else:
            deprecated = None
        index_symbols.append({
            "type": "record",
            "name": record.name,
            "ctype": record.base_ctype,
            "summary": utils.preprocess_docs(desc, repository.namespace, summary=True, plain=True),
            "deprecated": deprecated,
        })

        for ctor in record.constructors:
            if ctor.doc is not None:
                ctor_desc = ctor.doc.content
            else:
                ctor_desc = MISSING_DESCRIPTION
            if ctor.deprecated:
                (ctor_deprecated, _) = ctor.deprecated_since
            else:
                ctor_deprecated = None
            index_symbols.append({
                "type": "ctor",
                "name": ctor.name,
                "type_name": record.name,
                "ident": ctor.identifier,
                "summary": utils.preprocess_docs(ctor_desc, repository.namespace, summary=True, plain=True),
                "deprecated": ctor_deprecated,
            })

        for method in record.methods:
            if method.doc is not None:
                method_desc = method.doc.content
            else:
                method_desc = MISSING_DESCRIPTION
            index_symbols.append({
                "type": "method",
                "name": method.name,
                "type_name": record.name,
                "ident": method.identifier,
                "summary": utils.preprocess_docs(method_desc, repository.namespace, summary=True, plain=True),
            })

        for func in record.functions:
            if func.doc is not None:
                func_desc = func.doc.content
            else:
                func_desc = MISSING_DESCRIPTION
            index_symbols.append({
                "type": "type_func",
                "name": func.name,
                "type_name": record.name,
                "ident": func.identifier,
                "summary": utils.preprocess_docs(func_desc, repository.namespace, summary=True, plain=True),
            })


def _gen_unions(config, index, repository, symbols):
    index_symbols = index["symbols"]

    for union in symbols:
        if config.is_hidden(union.name):
            log.debug(f"Skipping hidden type {union.name}")
            continue
        if union.doc is not None:
            desc = union.doc.content
        else:
            desc = MISSING_DESCRIPTION
        if union.deprecated:
            (deprecated, _) = union.deprecated_since
        else:
            deprecated = None
        index_symbols.append({
            "type": "union",
            "name": union.name,
            "ctype": union.base_ctype,
            "summary": utils.preprocess_docs(desc, repository.namespace, summary=True, plain=True),
            "deprecated": deprecated,
        })

        for ctor in union.constructors:
            if ctor.doc is not None:
                ctor_desc = ctor.doc.content
            else:
                ctor_desc = MISSING_DESCRIPTION
            if ctor.deprecated:
                (ctor_deprecated, _) = ctor.deprecated_since
            else:
                ctor_deprecated = None
            index_symbols.append({
                "type": "ctor",
                "name": ctor.name,
                "type_name": union.name,
                "ident": ctor.identifier,
                "summary": utils.preprocess_docs(ctor_desc, repository.namespace, summary=True, plain=True),
                "deprecated": ctor_deprecated,
            })

        for method in union.methods:
            if method.doc is not None:
                method_desc = method.doc.content
            else:
                method_desc = MISSING_DESCRIPTION
            if method.deprecated:
                (method_deprecated, _) = method.deprecated_since
            else:
                method_deprecated = None
            index_symbols.append({
                "type": "method",
                "name": method.name,
                "type_name": union.name,
                "ident": method.identifier,
                "summary": utils.preprocess_docs(method_desc, repository.namespace, summary=True, plain=True),
                "deprecated": method_deprecated,
            })

        for func in union.functions:
            if func.doc is not None:
                func_desc = func.doc.content
            else:
                func_desc = MISSING_DESCRIPTION
            if func.deprecated:
                (func_deprecated, _) = func.deprecated_since
            else:
                func_deprecated = None
            index_symbols.append({
                "type": "type_func",
                "name": func.name,
                "type_name": union.name,
                "ident": func.identifier,
                "summary": utils.preprocess_docs(func_desc, repository.namespace, summary=True, plain=True),
                "deprecated": func_deprecated,
            })


def _gen_content_files(config, index, repository, content_dirs):
    index_symbols = index["symbols"]

    for file_name in config.content_files:
        src_file = utils.find_extra_content_file(content_dirs, file_name)

        src_data = ""
        with open(src_file, encoding='utf-8') as infile:
            source = []
            header = True
            title = None
            for line in infile:
                if header:
                    if line.startswith("Title: "):
                        title = line.replace("Title: ", "").strip()
                    if line == "\n":
                        header = False

                if not header:
                    source.append(line)
            src_data = "".join(source)

        if title is None:
            title = f"Untitled document '{file_name}'"

        index_symbols.append({
            "type": "content",
            "name": title,
            "href": file_name.replace(".md", ".html"),
            "summary": utils.preprocess_docs(src_data, repository.namespace, summary=True, plain=True),
        })


def gen_indices(config, repository, content_dirs, output_dir):
    namespace = repository.namespace

    symbols = {
        "aliases": sorted(namespace.get_aliases(), key=lambda alias: alias.name.lower()),
        "bitfields": sorted(namespace.get_bitfields(), key=lambda bitfield: bitfield.name.lower()),
        "callbacks": sorted(namespace.get_callbacks(), key=lambda callback: callback.name.lower()),
        "classes": sorted(namespace.get_classes(), key=lambda cls: cls.name.lower()),
        "constants": sorted(namespace.get_constants(), key=lambda const: const.name.lower()),
        "domains": sorted(namespace.get_error_domains(), key=lambda domain: domain.name.lower()),
        "enums": sorted(namespace.get_enumerations(), key=lambda enum: enum.name.lower()),
        "functions": sorted(namespace.get_functions(), key=lambda func: func.name.lower()),
        "function_macros": sorted(namespace.get_effective_function_macros(), key=lambda func: func.name.lower()),
        "interfaces": sorted(namespace.get_interfaces(), key=lambda interface: interface.name.lower()),
        "structs": sorted(namespace.get_effective_records(), key=lambda record: record.name.lower()),
        "unions": sorted(namespace.get_unions(), key=lambda union: union.name.lower()),
    }

    all_indices = {
        "aliases": _gen_aliases,
        "bitfields": _gen_bitfields,
        "callbacks": _gen_callbacks,
        "classes": _gen_classes,
        "constants": _gen_constants,
        "domains": _gen_domains,
        "enums": _gen_enums,
        "functions": _gen_functions,
        "function_macros": _gen_function_macros,
        "interfaces": _gen_interfaces,
        "structs": _gen_records,
        "unions": _gen_unions,
    }

    index = {
        "meta": {
            "ns": namespace.name,
            "version": namespace.version,
            "generator": "gi-docgen",
            "generator-version": core.version,
        },
        "symbols": [],
        "terms": {},
    }

    # Each section is isolated, so we run it into a thread pool
    for section in all_indices:
        generator = all_indices.get(section, None)
        if generator is None:
            log.error(f"No generator for section {section}")
            continue

        s = symbols.get(section, None)
        if s is None:
            log.debug(f"No symbols for section {section}")
            continue

        log.debug(f"Generating symbols for section {section}")
        generator(config, index, repository, s)

    _gen_content_files(config, index, repository, content_dirs)

    # Ensure iteration order is reproducible by sorting symbols by type/name,
    # and terms by key. This has no overhead since values are not copied.
    index["symbols"].sort(key=lambda s: (s["type"], s["name"]))
    index["terms"] = {}

    data = json.dumps(index, separators=(',', ':'))
    index_file = os.path.join(output_dir, "index.json")
    log.info(f"Creating index file for {namespace.name}-{namespace.version}: {index_file}")
    with open(index_file, "w", encoding="utf-8") as out:
        out.write(data)


def add_args(parser):
    parser.add_argument("--add-include-path", action="append", dest="include_paths", default=[],
                        help="include paths for other GIR files")
    parser.add_argument("-C", "--config", metavar="FILE", help="the configuration file")
    parser.add_argument("--content-dir", action="append", dest="content_dirs", default=[],
                        help="the base directories with the extra content")
    parser.add_argument("--dry-run", action="store_true", help="parses the GIR file without generating files")
    parser.add_argument("--output-dir", default=None, help="the output directory for the index files")
    parser.add_argument("infile", metavar="GIRFILE", type=argparse.FileType('r', encoding='UTF-8'),
                        default=sys.stdin, help="the GIR file to parse")


def run(options):
    log.info(f"Loading config file: {options.config}")

    conf = config.GIDocConfig(options.config)

    output_dir = options.output_dir or os.getcwd()
    log.info(f"Output directory: {output_dir}")

    content_dirs = options.content_dirs
    if content_dirs == []:
        content_dirs = [os.getcwd()]

    paths = []
    paths.extend(options.include_paths)
    paths.extend(utils.default_search_paths())
    log.debug(f"Search paths: {paths}")

    log.info("Parsing GIR file")
    parser = gir.GirParser(search_paths=paths)
    parser.parse(options.infile)

    if not options.dry_run:
        log.checkpoint()
        gen_indices(conf, parser.get_repository(), content_dirs, output_dir)

    return 0



# SPDX-FileCopyrightText: 2021 GNOME Foundation <https://gnome.org>
# SPDX-License-Identifier: Apache-2.0 OR GPL-3.0-or-later

import os
import re

toml_module = None
try:
    import tomllib as toml_lib
    toml_module = 'tomlib'
except ImportError:
    try:
        import tomli as toml_lib
        toml_module = 'tomli'
    except ImportError:
        import toml as toml_lib
        toml_module = 'toml'

from urllib.parse import urljoin
from packaging import version as packaging_version

from . import core, log, utils


class TomlConfig:
    """Wrapper class for TOML loading"""

    @staticmethod
    def load(toml):
        log.debug(f"Using TOML module: {toml_module}")
        if toml_module is None:
            log.error("No toml module found")
        elif toml_module in ['tomlib', 'tomli']:
            try:
                with open(toml, "rb") as f:
                    return toml_lib.load(f)
            except toml_lib.TOMLDecodeError as err:
                log.error(f"Invalid configuration file: {toml}: {err}")
        elif toml_module in ['toml']:
            try:
                return toml_lib.load(toml)
            except toml_lib.TomlDecodeError as err:
                log.error(f"Invalid configuration file: {toml}: {err}")


class GIDocConfig:
    """Load and represent the configuration for gidocgen"""
    def __init__(self, config_file=None):
        self._favicons = []
        self._config_file = config_file

        self._config = {}
        if self._config_file is not None:
            log.debug(f"Reading configuration file: {self._config_file}")
            self._config = TomlConfig.load(self._config_file)

    @property
    def library(self):
        return self._config.get('library', {})

    @property
    def extra(self):
        return self._config.get('extra', {})

    @property
    def theme(self):
        return self._config.get('theme', {})

    @property
    def check(self):
        return self._config.get('check', {})

    def get_templates_dir(self, default=None):
        return self.theme.get('templates_dir', default)

    def get_theme_name(self, default=None):
        return self.theme.get('name', default)

    def get_library_name(self, default=None):
        return self.library.get('name', default)

    def get_website_url(self, default=None):
        return self.library.get('website_url', default)

    def get_logo_url(self, default=None):
        return self.library.get('logo_url', default)

    def get_description(self, default=None):
        return self.library.get('description', default)

    @property
    def urlmap_file(self):
        return self.extra.get('urlmap_file')

    @property
    def urlmap_basename(self):
        if self.urlmap_file is not None:
            return os.path.basename(self.urlmap_file)
        return None

    @property
    def version(self):
        return self.library.get('version', 'Unknown')

    @property
    def authors(self):
        return self.library.get('authors', 'Unknown authors')

    @property
    def license(self):
        return self.library.get('license', 'All rights reserved')

    @property
    def website_url(self):
        return self.library.get('website_url', '')

    @property
    def docs_url(self):
        return self.library.get('docs_url', '')

    @property
    def browse_url(self):
        return self.library.get('browse_url', '')

    @property
    def logo_url(self):
        return self.library.get('logo_url', '')

    @property
    def description(self):
        return self.library.get('description', '')

    @property
    def dependencies(self):
        library = self._config.get('library', None)
        if library is None:
            return {}

        retval = {}
        dependencies = self._config.get('dependencies', {})
        for gir_name, dep in dependencies.items():
            res = {}
            res['name'] = dep.get('name', 'Unknown')
            res['description'] = dep.get('description', 'No description provided')
            res['docs_url'] = dep.get('docs_url', '#')
            retval[gir_name] = res
            log.debug(f"Found dependency {gir_name}: {res}")

        return retval

    @property
    def related(self):
        library = self._config.get('library', None)
        if library is None:
            return {}

        retval = {}
        dependencies = self._config.get('related', {})
        for gir_name, dep in dependencies.items():
            res = {}
            res['name'] = dep.get('name', 'Unknown')
            res['description'] = dep.get('description', 'No description provided')
            res['docs_url'] = dep.get('docs_url', '#')
            retval[gir_name] = res
            log.debug(f"Found related library {gir_name}: {res}")

        return retval

    @property
    def devhelp(self):
        return self.library.get('devhelp', False)

    @property
    def search_index(self):
        return self.library.get('search_index', False)

    @property
    def content_files(self):
        return self.extra.get('content_files', [])

    @property
    def content_images(self):
        return self.extra.get('content_images', [])

    @property
    def source_location_url(self):
        source_location = self._config.get('source-location', {})
        return source_location.get('base_url', '')

    @property
    def content_base_url(self):
        return self.extra.get('content_base_url')

    @property
    def file_format(self):
        source_location = self._config.get('source-location', {})
        return source_location.get('file_format', '{filename}#L{line}')

    @property
    def theme_name(self):
        return self.theme.get('name', '')

    @property
    def show_index_summary(self):
        return self.theme.get('show_index_summary', False)

    @property
    def show_class_hierarchy(self):
        # don't depend on dot(1) for ports builds; if around it will be used and
        # potentially junked by dpb(1) so just pretend it's not there by
        # checking for BSD_INSTALL_PROGRAM which is defined in bsd.port.mk(5)
        if os.getenv('BSD_INSTALL_PROGRAM') is not None:
            return False
        if utils.find_program('dot') is None:
            return False
        return self.theme.get('show_class_hierarchy', False)

    def source_link(self, *args):
        (filename, line) = args[0]
        base_url = self.source_location_url
        file_format = self.file_format
        endpoint = file_format.replace('{filename}', filename)
        endpoint = endpoint.replace('{line}', str(line))
        return urljoin(base_url, endpoint)

    def content_link(self, content_file):
        base_url = self.content_base_url
        return urljoin(base_url, content_file)

    @property
    def objects(self):
        return self._config.get('object', {})

    def match_object(self, name, match_key, category=None, key=None):
        def obj_matches(obj, name):
            n = obj.get('name')
            p = obj.get('pattern')
            if n is not None and n == name:
                return True
            elif p is not None and re.match(p, name):
                return True
            return False
        for obj in self.objects:
            if obj_matches(obj, name):
                if category is None:
                    return obj.get(match_key, False)
                else:
                    assert key is not None
                obj_category = obj.get(category)
                if obj_category is None:
                    return False
                for c in obj_category:
                    if obj_matches(c, key):
                        return c.get(match_key, False)
        return False

    def is_hidden(self, name, category=None, key=None):
        return self.match_object(name, 'hidden', category, key)

    def is_skipped(self, name, category=None, key=None):
        if self.is_hidden(name, category, key):
            return True
        return self.match_object(name, 'check_ignore', category, key)

    def is_unstable(self, version):
        if not version:
            return False
        cur_version = self.library.get('version')
        if cur_version is None:
            return False

        return packaging_version.parse(version) > packaging_version.parse(cur_version)

    @property
    def generator(self):
        return f"{core.version}"

    @property
    def favicons(self):
        if not self._favicons:
            self._favicons = [
                os.path.basename(p)
                for p in self.content_images
                if os.path.basename(p).startswith('favicon')
            ]

        return self._favicons

    @property
    def ignore_deprecated(self):
        return self.check.get('ignore_deprecated', False)


class GITemplateConfig:
    """Load and represent the template configuration"""
    def __init__(self, templates_dir, template_name):
        self._templates_dir = templates_dir
        self._template_name = template_name
        self._config_file = os.path.join(templates_dir, template_name, f"{template_name}.toml")

        self._config = {}
        log.debug(f"Reading template configuration file: {self._config_file}")
        self._config = TomlConfig.load(self._config_file)

    @property
    def name(self):
        metadata = self._config.get('metadata', {})
        return metadata.get('name', self._template_name)

    @property
    def css(self):
        css = self._config.get('css', {})
        return css.get('style', None)

    @property
    def extra_files(self):
        extra = self._config.get('extra_files', {})
        return extra.get('files', [])

    @property
    def templates(self):
        return self._config.get('templates', {})

    @property
    def class_template(self):
        return self.templates.get('class', 'class.html')

    @property
    def method_template(self):
        return self.templates.get('method', 'method.html')

    @property
    def class_method_template(self):
        return self.templates.get('class_method', 'class_method.html')

    @property
    def vfunc_template(self):
        return self.templates.get('vfunc', 'vfunc.html')

    @property
    def property_template(self):
        return self.templates.get('property', 'property.html')

    @property
    def signal_template(self):
        return self.templates.get('signal', 'signal.html')

    @property
    def type_func_template(self):
        return self.templates.get('type_func', 'type_func.html')

    @property
    def ctor_template(self):
        return self.templates.get('ctor', 'type_func.html')

    @property
    def func_template(self):
        return self.templates.get('function', 'function.html')

    @property
    def constant_template(self):
        return self.templates.get('constant', 'constant.html')

    @property
    def interface_template(self):
        return self.templates.get('interface', 'interface.html')

    @property
    def namespace_template(self):
        return self.templates.get('namespace', 'namespace.html')

    @property
    def content_template(self):
        return self.templates.get('content', 'content.html')

    @property
    def enum_template(self):
        return self.templates.get('enum', 'enum.html')

    @property
    def flags_template(self):
        return self.templates.get('flags', 'flags.html')

    @property
    def error_template(self):
        return self.templates.get('error', 'error.html')

    @property
    def record_template(self):
        return self.templates.get('record', 'record.html')

    @property
    def union_template(self):
        return self.templates.get('union', 'union.html')

    @property
    def alias_template(self):
        return self.templates.get('alias', 'alias.html')



# SPDX-FileCopyrightText: 2020 GNOME Foundation
# SPDX-License-Identifier: Apache-2.0 OR GPL-3.0-or-later

import typing as T

from .. import log


class Doc:
    """A documentation node, pointing to the source code"""
    def __init__(self, content: str, filename: str, line: int,
                 version: T.Optional[str] = None,
                 stability: T.Optional[str] = None):
        self.content = content
        self.filename = filename
        self.line = line
        self.version = version
        self.stability = stability

    def __str__(self):
        return self.content


class SourcePosition:
    """A location inside the source code"""
    def __init__(self, filename: str, line: int):
        self.filename = filename
        self.line = line

    def __str__(self):
        return f'{self.filename}:{self.line}'


class Attribute:
    """A user-defined annotation"""
    def __init__(self, name: str, value: T.Optional[str]):
        self.name = name
        self.value = value


class Include:
    """A GIR include"""
    def __init__(self, name: str, version: T.Optional[str] = None):
        self.name = name
        self.version = version

    def __str__(self):
        if self.version is not None:
            return f"{self.name}-{self.version}"
        return f"{self.name}"

    def girfile(self) -> str:
        if self.version is not None:
            return f"{self.name}-{self.version}.gir"
        return f"{self.name}.gir"


class Info:
    """Base information for most types"""
    def __init__(self, introspectable: bool = True, deprecated: T.Optional[str] = None,
                 deprecated_version: T.Optional[str] = None, version: T.Optional[str] = None,
                 stability: T.Optional[str] = None):
        self.introspectable = introspectable
        self.deprecated_msg = deprecated
        self.deprecated_version = deprecated_version
        self.deprecated = deprecated is not None or deprecated_version is not None
        self.version = version
        self.stability = stability
        self.attributes: T.Mapping[str, T.Optional[str]] = {}
        self.doc: T.Optional[Doc] = None
        self.source_position: T.Optional[SourcePosition] = None

    def add_attribute(self, name: str, value: T.Optional[str] = None) -> None:
        self.attributes[name] = value


class GIRElement:
    """Base type for elements inside the GIR"""
    def __init__(self, name: T.Optional[str] = None, namespace: T.Optional[str] = None):
        self.name = name
        self.namespace = namespace
        if self.namespace is None:
            if self.name is not None and '.' in self.name:
                self.namespace = self.name.split('.')[0]
        self.info = Info()

    def set_introspectable(self, introspectable: bool) -> None:
        """Set whether the symbol is introspectable"""
        self.info.introspectable = introspectable

    @property
    def introspectable(self):
        return self.info.introspectable

    def set_version(self, version: str) -> None:
        """Set the version of the symbol"""
        self.info.version = version

    def set_stability(self, stability: str) -> None:
        """Set the stability of the symbol"""
        self.info.stability = stability

    @property
    def stability(self):
        return self.info.stability

    def set_doc(self, doc: Doc) -> None:
        """Set the documentation for the element"""
        self.info.doc = doc

    @property
    def doc(self):
        return self.info.doc

    def set_source_position(self, pos: SourcePosition) -> None:
        """Set the position in the source code for the element"""
        self.info.source_position = pos

    @property
    def source_position(self) -> T.Optional[T.Tuple[str, int]]:
        if self.info.source_position is None:
            return None
        return self.info.source_position.filename, self.info.source_position.line

    def set_deprecated(self, doc: T.Optional[str] = None,
                       since_version: T.Optional[str] = None) -> None:
        """Set the deprecation annotations for the element"""
        self.info.deprecated_msg = doc
        self.info.deprecated_version = since_version
        self.info.deprecated = True

    def set_attributes(self, attrs: T.Mapping[str, T.Optional[str]]) -> None:
        """Add an annotation to the symbol"""
        for name, value in attrs.items():
            self.info.add_attribute(name, value)

    @property
    def attributes(self) -> T.Mapping[str, T.Optional[str]]:
        return self.info.attributes

    @property
    def available_since(self) -> T.Optional[str]:
        return self.info.version

    @property
    def deprecated(self) -> bool:
        return self.info.deprecated

    @property
    def deprecated_since(self) -> T.Optional[T.Tuple[str, str]]:
        if not self.info.deprecated:
            return None
        version = self.info.deprecated_version
        message = self.info.deprecated_msg
        if version is None:
            version = "Unknown"
        if message is None:
            message = "Please do not use it in newly written code"
        return (version, message)


class Type(GIRElement):
    """Base class for all Type nodes"""
    def __init__(self, name: str, ctype: T.Optional[str] = None, namespace: T.Optional[str] = None, is_fundamental: bool = False):
        super().__init__(name=name, namespace=namespace)
        self.ctype = ctype
        self.is_fundamental = is_fundamental

    def __eq__(self, other):
        if isinstance(other, Type):
            if self.namespace is not None:
                return self.namespace == other.namespace and self.name == other.name
            elif self.ctype is not None:
                return self.name == other.name and self.ctype == other.ctype
            else:
                return self.name == other.name
        elif isinstance(other, str):
            return self.name == other
        else:
            return False

    def __cmp__(self, other):
        if self.ctype is not None:
            return self.name == other.name and self.ctype == other.ctype
        return self.name == other.name

    def __repr__(self):
        return f"Type({self.fqtn}, {self.ctype})"

    @property
    def resolved(self):
        return self.ctype is not None

    @property
    def base_ctype(self):
        if self.ctype is None:
            return None
        return self.ctype.replace('*', '')

    @property
    def fqtn(self):
        if self.is_fundamental:
            return self.name
        elif '.' in self.name:
            return self.name
        elif self.namespace is not None:
            return f"{self.namespace}.{self.name}"
        else:
            return None


class ArrayType(GIRElement):
    """Base class for Array nodes"""
    def __init__(self, name: T.Optional[str], value_type: Type, ctype: T.Optional[str] = None, zero_terminated: bool = False,
                 fixed_size: int = -1, length: int = -1):
        super().__init__(name)
        self.ctype = ctype
        self.zero_terminated = zero_terminated
        self.fixed_size = fixed_size
        self.length = length
        self.value_type = value_type
        self.is_fundamental = False


class ListType(GIRElement):
    """Type class for List nodes"""
    def __init__(self, name: str, value_type: Type, ctype: T.Optional[str] = None):
        super().__init__(name)
        self.ctype = ctype
        self.value_type = value_type
        self.is_fundamental = False


class MapType(GIRElement):
    """Type class for Map nodes"""
    def __init__(self, name: str, key_type: Type, value_type: Type, ctype: T.Optional[str] = None):
        super().__init__(name)
        self.ctype = ctype
        self.key_type = key_type
        self.value_type = value_type
        self.is_fundamental = False


class GType:
    """Base class for GType information"""
    def __init__(self, type_name: str, get_type: str, type_struct: T.Optional[str] = None):
        self.type_name = type_name
        self.get_type = get_type
        self.type_struct = type_struct


class VoidType(Type):
    def __init__(self):
        super().__init__(name='none', ctype='void')

    def __str__(self):
        return "void"


class VarArgs(Type):
    def __init__(self):
        super().__init__(name='none', ctype='')

    def __str__(self):
        return "..."


class Alias(Type):
    """Alias to a Type"""
    def __init__(self, name: str, namespace: str, ctype: str, target: Type):
        super().__init__(name=name, ctype=ctype, namespace=namespace)
        self.target = target


class Constant(Type):
    """A constant"""
    def __init__(self, name: str, namespace: str, ctype: str, target: Type, value: str):
        super().__init__(name=name, ctype=ctype, namespace=namespace)
        self.target = target
        self.value = value


class Parameter(GIRElement):
    """A callable parameter"""
    def __init__(self, name: str, direction: str, transfer: str, target: T.Optional[Type] = None, caller_allocates: bool = False,
                 optional: bool = False, nullable: bool = False, closure: int = -1, destroy: int = -1,
                 scope: T.Optional[str] = None):
        super().__init__(name)
        self.direction = direction
        self.transfer = transfer
        self.caller_allocates = caller_allocates
        self.optional = optional
        self.nullable = nullable
        self.scope = scope
        self.closure = closure
        self.destroy = destroy
        if target is None:
            self.target: Type = VoidType()
        else:
            self.target = target


class ReturnValue(GIRElement):
    """A callable's return value"""
    def __init__(self, transfer: str, target: Type, nullable: bool = False,
                 closure: int = -1, destroy: int = -1,
                 scope: T.Optional[str] = None):
        super().__init__()
        self.transfer = transfer
        self.nullable = nullable
        self.scope = scope
        self.closure = closure
        self.destroy = destroy
        if target is None:
            self.target: Type = VoidType()
        else:
            self.target = target


class Callable(GIRElement):
    """A callable symbol: function, method, function-macro, ..."""
    def __init__(self, name: str, namespace: T.Optional[str], identifier: T.Optional[str], throws: bool = False,
                 inline: bool = False):
        super().__init__(name=name, namespace=namespace)
        self.identifier = identifier
        self.parameters: T.List[Parameter] = []
        self.return_value: T.Optional[ReturnValue] = None
        self.throws: bool = throws
        self.inline: bool = inline
        self.moved_to: T.Optional[str] = None
        self.shadows: T.Optional[str] = None
        self.shadowed_by: T.Optional[str] = None
        self.async_func: T.Optional[str] = None
        self.sync_func: T.Optional[str] = None
        self.finish_func: T.Optioal[str] = None

    def add_parameter(self, param: Parameter) -> None:
        self.parameters.append(param)

    def set_parameters(self, params: T.List[Parameter]) -> None:
        self.parameters.extend(params)

    def set_return_value(self, res: ReturnValue) -> None:
        self.return_value = res

    def set_shadows(self, func: str) -> None:
        self.shadows = func

    def set_shadowed_by(self, func: str) -> None:
        self.shadowed_by = func

    def set_moved_to(self, func: str) -> None:
        self.moved_to = func

    def set_async_func(self, func: T.Optional[str]) -> None:
        self.async_func = func

    def set_sync_func(self, func: T.Optional[str]) -> None:
        self.sync_func = func

    def set_finish_func(self, func: T.Optional[str]) -> None:
        self.finish_func = func

    def __contains__(self, param):
        if isinstance(param, str):
            for p in self.parameters:
                if p.name == param:
                    return True
        elif isinstance(param, Parameter):
            return param in self.parameters
        elif isinstance(param, ReturnValue):
            return param == self.return_value
        return False


class FunctionMacro(Callable):
    def __init__(self, name: str, namespace: T.Optional[str], identifier: str):
        super().__init__(name, namespace, identifier)


class Function(Callable):
    def __init__(self, name: str, namespace: T.Optional[str], identifier: str, throws: bool = False,
                 inline: bool = False):
        super().__init__(name, namespace, identifier, throws, inline)


class Method(Callable):
    def __init__(self, name: str, identifier: str, instance_param: Parameter, throws: bool = False,
                 set_property: T.Optional[str] = None, get_property: T.Optional[str] = None,
                 inline: bool = False):
        super().__init__(name, None, identifier, throws, inline)
        self.instance_param = instance_param
        self.set_property = set_property
        self.get_property = get_property

    def __contains__(self, param):
        if isinstance(param, Parameter) and param == self.instance_param:
            return True
        return super().__contains__(self, param)


class VirtualMethod(Callable):
    def __init__(self, name: str, identifier: str, invoker: str, instance_param: T.Optional[Parameter],
                 static: bool = False, throws: bool = False):
        super().__init__(name, None, identifier, throws)
        self.instance_param = instance_param
        self.invoker = invoker
        self.static = static

    def __contains__(self, param):
        if isinstance(param, Parameter) and param == self.instance_param:
            return True
        return super().__contains__(self, param)


class Callback(Callable):
    def __init__(self, name: str, namespace: str, ctype: T.Optional[str], throws: bool = False):
        super().__init__(name=name, namespace=namespace, identifier=None, throws=throws)
        self.ctype = ctype
        self.is_fundamental = False

    @property
    def base_ctype(self):
        if self.ctype is None:
            return None
        return self.ctype.replace('*', '')


class Member(GIRElement):
    """A member in an enumeration, error domain, or bitfield"""
    def __init__(self, name: str, value: str, identifier: str, nick: str):
        super().__init__(name)
        self.value = value
        self.identifier = identifier
        self.nick = nick


class Enumeration(Type):
    """An enumeration type"""
    def __init__(self, name: str, namespace: str, ctype: str, gtype: T.Optional[GType]):
        super().__init__(name=name, ctype=ctype, namespace=namespace)
        self.gtype = gtype
        self.members: T.List[Member] = []
        self.functions: T.List[Function] = []

    def add_member(self, member: Member) -> None:
        self.members.append(member)

    def add_function(self, function: Function) -> None:
        self.functions.append(function)

    def set_members(self, members: T.List[Member]) -> None:
        self.members.extend(members)

    def set_functions(self, functions: T.List[Function]) -> None:
        self.functions.extend(functions)

    def __contains__(self, member):
        if isinstance(member, Member):
            return member in self.members
        return False

    def __iter__(self):
        for member in self.members:
            yield member


class BitField(Enumeration):
    """An enumeration type of bit masks"""
    def __init__(self, name: str, namespace: str, ctype: str, gtype: T.Optional[GType]):
        super().__init__(name, namespace, ctype, gtype)


class ErrorDomain(Enumeration):
    """An error domain for GError"""
    def __init__(self, name: str, namespace: str, ctype: str, gtype: T.Optional[GType], domain: str):
        super().__init__(name, namespace, ctype, gtype)
        self.domain = domain


class Property(GIRElement):
    def __init__(self, name: str, transfer: str, target: Type, writable: bool = True, readable: bool = True,
                 construct: bool = False, construct_only: bool = False, setter: T.Optional[str] = None,
                 getter: T.Optional[str] = None, default_value: T.Optional[str] = None):
        super().__init__(name)
        self.transfer = transfer
        self.writable = writable
        self.readable = readable
        self.construct = construct
        self.construct_only = construct_only
        self.target = target
        self.setter = setter
        self.getter = getter
        self.default_value = default_value


class Signal(GIRElement):
    def __init__(self, name: str, detailed: bool, when: str, action: bool = False, no_hooks: bool = False, no_recurse: bool = False):
        super().__init__(name)
        self.detailed = detailed
        self.when = when
        self.action = action
        self.no_hooks = no_hooks
        self.no_recurse = no_recurse
        self.parameters: T.List[Parameter] = []
        self.return_value: T.Optional[ReturnValue] = None

    def set_parameters(self, params: T.List[Parameter]) -> None:
        self.parameters.extend(params)

    def set_return_value(self, res: ReturnValue) -> None:
        self.return_value = res


class Field(GIRElement):
    """A field in a struct or union"""
    def __init__(self, name: str, target: Type, writable: bool, readable: bool, private: bool = False, bits: int = 0):
        super().__init__(name)
        self.target = target
        self.writable = writable
        self.readable = readable
        self.private = private
        self.bits = bits


class Interface(Type):
    def __init__(self, name: str, namespace: str, ctype: str, symbol_prefix: str, gtype: GType):
        super().__init__(name=name, ctype=ctype, namespace=namespace)
        self.symbol_prefix = symbol_prefix
        self.gtype = gtype
        self.methods: T.List[Method] = []
        self.virtual_methods: T.List[VirtualMethod] = []
        self.properties: T.Mapping[str, Property] = {}
        self.signals: T.Mapping[str, Signal] = {}
        self.functions: T.List[Function] = []
        self.fields: T.List[Field] = []
        self.prerequisite: T.Optional[str] = None
        self.implementations: T.List[Type] = []

    @property
    def type_struct(self) -> T.Optional[str]:
        if self.gtype is not None:
            return self.gtype.type_struct
        return self.ctype

    @property
    def type_func(self) -> str:
        return self.gtype.get_type

    def set_methods(self, methods: T.List[Method]) -> None:
        self.methods.extend(methods)

    def set_virtual_methods(self, methods: T.List[VirtualMethod]) -> None:
        self.virtual_methods.extend(methods)

    def set_properties(self, properties: T.List[Property]) -> None:
        for p in properties:
            self.properties[p.name] = p

    def set_signals(self, signals: T.List[Signal]) -> None:
        for s in signals:
            self.signals[s.name] = s

    def set_functions(self, functions: T.List[Function]) -> None:
        self.functions.extend(functions)

    def set_fields(self, fields: T.List[Field]) -> None:
        self.fields.extend(fields)

    def set_prerequisite(self, prerequisite: str) -> None:
        self.prerequisite = prerequisite


class Class(Type):
    def __init__(self, name: str, namespace: str, ctype: str, symbol_prefix: str,
                 gtype: GType, parent: T.Optional[Type] = None,
                 abstract: bool = False, fundamental: bool = False,
                 ref_func: T.Optional[str] = None, unref_func: T.Optional[str] = None):
        super().__init__(name=name, ctype=ctype, namespace=namespace)
        self.symbol_prefix = symbol_prefix
        self.parent = parent
        self.abstract = abstract
        self.fundamental = fundamental
        self.ref_func = ref_func
        self.unref_func = unref_func
        self.gtype = gtype
        self.ancestors: T.List[Type] = []
        self.implements: T.List[Type] = []
        self.constructors: T.List[Function] = []
        self.methods: T.List[Method] = []
        self.virtual_methods: T.List[VirtualMethod] = []
        self.properties: T.Mapping[str, Property] = {}
        self.signals: T.Mapping[str, Signal] = {}
        self.functions: T.List[Function] = []
        self.fields: T.List[Field] = []
        self.callbacks: T.List[Callback] = []
        self.descendants: T.List[Type] = []

    @property
    def type_struct(self) -> T.Optional[str]:
        if self.gtype is not None:
            return self.gtype.type_struct
        return None

    @property
    def type_func(self) -> T.Optional[str]:
        if self.gtype is not None:
            return self.gtype.get_type
        return self.ctype

    def set_constructors(self, ctors: T.List[Function]) -> None:
        self.constructors.extend(ctors)

    def set_methods(self, methods: T.List[Method]) -> None:
        self.methods.extend(methods)

    def set_virtual_methods(self, methods: T.List[VirtualMethod]) -> None:
        self.virtual_methods.extend(methods)

    def set_properties(self, properties: T.List[Property]) -> None:
        for p in properties:
            self.properties[p.name] = p

    def set_signals(self, signals: T.List[Signal]) -> None:
        for s in signals:
            self.signals[s.name] = s

    def set_functions(self, functions: T.List[Function]) -> None:
        self.functions.extend(functions)

    def set_implements(self, ifaces: T.List[Type]) -> None:
        self.implements.extend(ifaces)

    def set_fields(self, fields: T.List[Field]) -> None:
        self.fields.extend(fields)


class Boxed(Type):
    def __init__(self, name: str, namespace: str, symbol_prefix: str, gtype: GType):
        super().__init__(name=name, ctype=None, namespace=namespace)
        self.symbol_prefix = symbol_prefix
        self.gtype = gtype
        self.functions: T.List[Function] = []

    def set_functions(self, functions: T.List[Function]) -> None:
        self.functions.extend(functions)


class Record(Type):
    def __init__(self, name: str, namespace: str, ctype: str, symbol_prefix: str,
                 gtype: T.Optional[GType] = None, struct_for: T.Optional[str] = None,
                 disguised: bool = False):
        super().__init__(name=name, ctype=ctype, namespace=namespace)
        self.symbol_prefix = symbol_prefix
        self.gtype = gtype
        self.struct_for = struct_for
        self.disguised = disguised
        self.constructors: T.List[Function] = []
        self.methods: T.List[Method] = []
        self.functions: T.List[Function] = []
        self.fields: T.List[Field] = []

    @property
    def type_struct(self) -> T.Optional[str]:
        if self.gtype is not None:
            return self.gtype.type_struct
        return self.ctype

    @property
    def type_func(self) -> T.Optional[str]:
        if self.gtype is not None:
            return self.gtype.get_type
        return None

    def set_constructors(self, ctors: T.List[Function]) -> None:
        self.constructors.extend(ctors)

    def set_methods(self, methods: T.List[Method]) -> None:
        self.methods.extend(methods)

    def set_functions(self, functions: T.List[Function]) -> None:
        self.functions.extend(functions)

    def set_fields(self, fields: T.List[Field]) -> None:
        self.fields.extend(fields)


class Union(Type):
    def __init__(self, name: str, namespace: str, ctype: str, symbol_prefix: str, gtype: T.Optional[GType]):
        super().__init__(name=name, ctype=ctype, namespace=namespace)
        self.symbol_prefix = symbol_prefix
        self.gtype = gtype
        self.constructors: T.List[Function] = []
        self.methods: T.List[Method] = []
        self.functions: T.List[Function] = []
        self.fields: T.List[Field] = []

    @property
    def type_struct(self) -> T.Optional[str]:
        if self.gtype is not None:
            return self.gtype.type_struct
        return self.ctype

    @property
    def type_func(self) -> T.Optional[str]:
        if self.gtype is not None:
            return self.gtype.get_type
        return None

    def set_constructors(self, ctors: T.List[Function]) -> None:
        self.constructors.extend(ctors)

    def set_methods(self, methods: T.List[Method]) -> None:
        self.methods.extend(methods)

    def set_functions(self, functions: T.List[Function]) -> None:
        self.functions.extend(functions)

    def set_fields(self, fields: T.List[Field]) -> None:
        self.fields.extend(fields)


class Namespace:
    def __init__(self, name: str, version: str, identifier_prefix: T.List[str] = [], symbol_prefix: T.List[str] = []):
        self.name = name
        self.version = version

        self._shared_libraries: T.List[str] = []

        self._aliases: T.Mapping[str, Alias] = {}
        self._bitfields: T.Mapping[str, BitField] = {}
        self._boxeds: T.Mapping[str, Boxed] = {}
        self._callbacks: T.Mapping[str, Callback] = {}
        self._classes: T.Mapping[str, Class] = {}
        self._constants: T.Mapping[str, Constant] = {}
        self._enumerations: T.Mapping[str, Enumeration] = {}
        self._error_domains: T.Mapping[str, ErrorDomain] = {}
        self._functions: T.Mapping[str, Function] = {}
        self._function_macros: T.Mapping[str, FunctionMacro] = {}
        self._interfaces: T.Mapping[str, Interface] = {}
        self._records: T.Mapping[str, Record] = {}
        self._unions: T.Mapping[str, Union] = {}

        self._symbols: T.Mapping[str, Type] = {}
        self.repository: T.Optional[Repository] = None

        if identifier_prefix:
            self.identifier_prefix = identifier_prefix
        else:
            self.identifier_prefix = [self.name]
        if symbol_prefix:
            self.symbol_prefix = symbol_prefix
        else:
            self.symbol_prefix = [self.name.lower()]

    def __str__(self):
        return f"{self.name}-{self.version}"

    def add_shared_libraries(self, libs: T.List[str]) -> None:
        self._shared_libraries.extend(libs)

    def get_shared_libraries(self) -> T.List[str]:
        return self._shared_libraries

    def add_alias(self, alias: Alias) -> None:
        self._aliases[alias.name] = alias

    def add_enumeration(self, enum: Enumeration) -> None:
        self._enumerations[enum.name] = enum

    def add_error_domain(self, domain: ErrorDomain) -> None:
        self._error_domains[domain.name] = domain

    def add_class(self, cls: Class) -> None:
        self._classes[cls.name] = cls

    def add_constant(self, constant: Constant) -> None:
        self._constants[constant.name] = constant

    def add_interface(self, interface: Interface) -> None:
        self._interfaces[interface.name] = interface

    def add_boxed(self, boxed: Boxed) -> None:
        self._boxeds[boxed.name] = boxed

    def add_record(self, record: Record) -> None:
        self._records[record.name] = record

    def add_union(self, union: Union) -> None:
        self._unions[union.name] = union

    def add_function(self, function: Function) -> None:
        self._functions[function.name] = function

    def add_bitfield(self, bitfield: BitField) -> None:
        self._bitfields[bitfield.name] = bitfield

    def add_function_macro(self, function: FunctionMacro) -> None:
        self._function_macros[function.name] = function

    def add_callback(self, callback: Callback) -> None:
        self._callbacks[callback.name] = callback

    def get_classes(self) -> T.List[Class]:
        return self._classes.values()

    def get_constants(self) -> T.List[Constant]:
        return self._constants.values()

    def get_enumerations(self) -> T.List[Enumeration]:
        return self._enumerations.values()

    def get_error_domains(self) -> T.List[ErrorDomain]:
        return self._error_domains.values()

    def get_aliases(self) -> T.List[Alias]:
        return self._aliases.values()

    def get_interfaces(self) -> T.List[Interface]:
        return self._interfaces.values()

    def get_boxeds(self) -> T.List[Boxed]:
        return self._boxeds.values()

    def get_records(self) -> T.List[Record]:
        return self._records.values()

    def get_effective_records(self) -> T.List[Record]:
        def is_effective(r):
            if "Private" in r.name and r.disguised:
                return False
            if r.struct_for is not None:
                return False
            return True

        return [x for x in self._records.values() if is_effective(x)]

    def get_unions(self) -> T.List[Union]:
        return self._unions.values()

    def get_functions(self) -> T.List[Function]:
        return self._functions.values()

    def get_bitfields(self) -> T.List[BitField]:
        return self._bitfields.values()

    def get_function_macros(self) -> T.List[FunctionMacro]:
        return self._function_macros.values()

    def get_effective_function_macros(self) -> T.List[FunctionMacro]:
        def is_effective(f, ns):
            # Lower-case identifiers are an automatic pass
            if f.name.islower():
                return True
            # Try to eliminate the GObject type macros from the pool
            t = f.name.split('_')
            # Skip "is-a" macros
            if 'IS' in t:
                return False
            # Skip "get class/iface" macros
            if 'GET' in t:
                return False
            # Re-assemble into what most likely is a type name
            s = "".join([x.capitalize() if len(x) > 2 else x for x in t])
            # Skip "cast" macros
            if ns.find_class(s) is not None:
                return False
            if ns.find_interface(s) is not None:
                return False
            if ns.find_record(s) is not None:
                return False
            # Anything that survived at this point is likely a valid function
            # macro
            return True

        return [x for x in self._function_macros.values() if is_effective(x, self)]

    def get_callbacks(self) -> T.List[Callback]:
        return self._callbacks.values()

    def find_class(self, cls: str) -> T.Optional[Class]:
        return self._classes.get(cls)

    def find_record(self, record: str) -> T.Optional[Record]:
        return self._records.get(record)

    def find_interface(self, iface: str) -> T.Optional[Interface]:
        return self._interfaces.get(iface)

    def find_union(self, union: str) -> T.Optional[Union]:
        return self._unions.get(union)

    def find_enumeration(self, enum: str) -> T.Optional[Enumeration]:
        return self._enumerations.get(enum)

    def find_bitfield(self, bitfield: str) -> T.Optional[BitField]:
        return self._bitfields.get(bitfield)

    def find_error_domain(self, domain: str) -> T.Optional[ErrorDomain]:
        return self._error_domains.get(domain)

    def find_alias(self, alias: str) -> T.Optional[Alias]:
        return self._aliases.get(alias)

    def find_function(self, func: str) -> T.Optional[Function]:
        if func in self._functions:
            return self._functions.get(func)
        if func in self._function_macros:
            return self._function_macros.get(func)
        return None

    def find_real_type(self, name: str) -> T.Optional[Type]:
        if name in self._aliases:
            return self._aliases[name]
        if name in self._bitfields:
            return self._bitfields[name]
        if name in self._callbacks:
            return self._callbacks[name]
        if name in self._constants:
            return self._constants[name]
        if name in self._enumerations:
            return self._enumerations[name]
        if name in self._error_domains:
            return self._error_domains[name]
        if name in self._classes:
            return self._classes[name]
        if name in self._interfaces:
            return self._interfaces[name]
        if name in self._records:
            return self._records[name]
        if name in self._unions:
            return self._unions[name]
        return None

    def find_symbol(self, name: str) -> T.Optional[Type]:
        return self._symbols.get(name)

    def find_prerequisite_type(self, name: str) -> T.Optional[Type]:
        if name in self._classes:
            return self._classes[name]
        if name in self._interfaces:
            return self._interfaces[name]
        return None


class Repository:
    def __init__(self):
        self.includes: T.Mapping[str, Repository] = {}
        self.packages: T.List[str] = []
        self.c_includes: T.List[str] = []
        self.types: T.Mapping[str, T.List[Type]] = {}
        self._namespaces: T.List[Namespace] = []
        self.girfile: T.Optional[str] = None

    def add_namespace(self, ns: Namespace) -> None:
        self._namespaces.append(ns)
        ns.repository = self

    def get_namespace(self, ns: str) -> T.Optional[Namespace]:
        for namespace in self._namespaces:
            if namespace.name == ns:
                return namespace
        return None

    def find_included_namespace(self, ns: str) -> T.Optional[Namespace]:
        for repo_name in self.includes:
            repo = self.includes[repo_name]
            if repo.namespace.name == ns:
                return repo.namespace
        return None

    def _lookup_type(self, name: str) -> T.Optional[Type]:
        types = self.types.get(name)
        if types is None:
            return None
        for t in types:
            if t.resolved:
                return t
        return types[0]

    def resolve_empty_ctypes(self, seen_types: T.Mapping[str, T.List[Type]]) -> None:
        for fqtn in seen_types:
            types = seen_types[fqtn]
            resolved_types = [t for t in types if t.resolved]
            if len(resolved_types) == 0:
                ns, name = fqtn.split('.', 1)
                backstop = f"{self.namespace.identifier_prefix[0]}{name}"
                resolved_types.append(Type(fqtn, backstop))
            self.types[fqtn] = resolved_types
            log.debug(f"Type: {fqtn}: {resolved_types}")

    def resolve_interface_requires(self) -> None:
        def find_prerequisite_type(includes, ns, name):
            repository = includes.get(ns)
            if repository is None:
                return None
            prereq = repository.namespace.find_prerequisite_type(name)
            # If the prerequisite type is unqualified, then we qualify it here
            if '.' not in prereq.name:
                prereq.name = f"{repository.namespace.name}.{prereq.name}"
            return prereq

        ifaces = self.namespace.get_interfaces()
        for iface in ifaces:
            if iface.prerequisite is None:
                continue
            prerequisite = None
            if '.' in iface.prerequisite.name:
                ns, name = iface.prerequisite.name.split('.', 1)
                if ns == self.namespace.name:
                    prerequisite = self.namespace.find_prerequisite_type(name)
                else:
                    prerequisite = find_prerequisite_type(self.includes, ns, name)
            else:
                prerequisite = self.namespace.find_prerequisite_type(iface.prerequisite.name)
            if prerequisite is not None:
                if prerequisite.ctype is None:
                    if '.' not in prerequisite.name:
                        name = f"{self.namespace.name}.{prerequisite.name}"
                    else:
                        name = prerequisite.name
                    t = self._lookup_type(name)
                    if t is not None:
                        prerequisite.ctype = t.ctype
                    else:
                        # This is kind of a kludge, but apparently we can get into
                        # class definitions missing a c:type; if that happens, we
                        # take the identifier prefix of the namespace and append the
                        # class name, because that's the inverse of how g-ir-scanner
                        # determines the class name
                        prerequisite.ctype = f"{self.namespace.identifier_prefix[0]}{prerequisite.name}"
                iface.prerequisite = prerequisite
                log.debug(f"Prerequisite type for interface {iface}: {iface.prerequisite}")

    def resolve_class_ctype(self) -> None:
        classes = self.namespace.get_classes()
        for cls in classes:
            if cls.ctype is None:
                if '.' not in cls.name:
                    name = f"{self.namespace.name}.{cls.name}"
                else:
                    name = cls.name
                t = self._lookup_type(name)
                if t is not None:
                    cls.ctype = t.base_ctype
                else:
                    # This is kind of a kludge, but apparently we can get into
                    # class definitions missing a c:type; if that happens, we
                    # take the identifier prefix of the namespace and append the
                    # class name, because that's the inverse of how g-ir-scanner
                    # determines the class name
                    cls.ctype = f"{self.namespace.identifier_prefix[0]}{cls.name}"
                log.debug(f"Updated C type for {cls}")

    def resolve_class_implements(self) -> None:
        def find_interface_type(includes, ns, name):
            repository = includes.get(ns)
            if repository is None:
                return None
            iface = repository.namespace.find_interface(name)
            # If the interface type is unqualified, then we qualify it here
            if '.' not in iface.name:
                iface.name = f"{repository.namespace.name}.{iface.name}"
            return iface

        classes = self.namespace.get_classes()
        for cls in classes:
            if cls.implements is None:
                continue
            implements = cls.implements
            cls.implements = []
            for iface in implements:
                if '.' in iface.name:
                    ns, name = iface.name.split('.', 1)
                    if ns == self.namespace.name:
                        iface_type = self.namespace.find_interface(name)
                    else:
                        iface_type = find_interface_type(self.includes, ns, name)
                else:
                    iface_type = self.namespace.find_interface(iface.name)
                if iface_type is not None:
                    if iface_type.ctype is None:
                        t = self._lookup_type(iface_type.name)
                        iface_type.ctype = t.ctype
                    cls.implements.append(iface_type)
            log.debug(f"Interfaces implemented by {cls}: {cls.implements}")

    def resolve_class_ancestors(self) -> None:
        def find_parent_class(includes, ns, name):
            repository = includes.get(ns)
            if repository is None:
                return None
            parent_class = repository.namespace.find_class(name)
            # If the parent type is unqualified, then we qualify it here
            if '.' not in parent_class.name:
                parent_class.name = f"{repository.namespace.name}.{parent_class.name}"
            return parent_class

        classes = self.namespace.get_classes()
        for cls in classes:
            if cls.parent is None:
                continue
            ancestors = []
            parent = cls.parent
            while parent is not None:
                if '.' in parent.name:
                    ns, name = parent.name.split('.')
                    if ns == self.namespace.name:
                        real_parent = self.namespace.find_class(name)
                    else:
                        real_parent = find_parent_class(self.includes, ns, name)
                else:
                    real_parent = self.namespace.find_class(parent.name)
                if real_parent is None:
                    break
                if real_parent.parent is not None and real_parent.parent.name == parent.name:
                    log.warning(f"Found a loop in the ancestors for {cls}: {real_parent} matches {parent}")
                    break
                if real_parent.ctype is None:
                    log.debug(f"Looking up C type for {parent.fqtn}")
                    t = self._lookup_type(parent.name)
                    real_parent.ctype = t.ctype
                log.debug(f"Adding ancestor {real_parent} for {cls}")
                ancestors.append(real_parent)
                parent = real_parent.parent
            cls.ancestors = ancestors
            cls.parent = ancestors[0]
            log.debug(f"Ancestors for {cls}: parent: {cls.parent}, ancestors: {cls.ancestors}")

    def resolve_class_descendants(self) -> None:
        seen_parents = {}
        for cls in self.namespace.get_classes():
            if cls.parent is not None:
                seen_parents.setdefault(cls.parent.name, []).append(cls)
        for name, descendants in seen_parents.items():
            if name in self.namespace._classes:
                self.namespace._classes[name].descendants = descendants

    def resolve_moved_to(self) -> None:
        functions = list(self.namespace.get_functions())
        old_len = len(functions)
        for func in functions[:]:
            if func.moved_to is None:
                continue
            moved_type, moved_func_name = func.moved_to.split('.')
            real_type = self.namespace.find_real_type(moved_type)
            if real_type is None:
                continue
            self.namespace._functions.pop(func.name)    # XXX: Add accessor
        new_len = len(self.namespace._functions)
        diff = old_len - new_len
        log.debug(f"Removed {old_len} - {new_len} functions: {diff}")

    def resolve_symbols(self) -> None:
        symbols: T.Mapping[str, Type] = {}
        for func in self.namespace.get_functions():
            symbols[func.identifier] = func
        for func in self.namespace.get_function_macros():
            symbols[func.identifier] = func
        for cls in self.namespace.get_classes():
            for m in cls.constructors:
                symbols[m.identifier] = cls
            for m in cls.methods:
                symbols[m.identifier] = cls
            for m in cls.functions:
                symbols[m.identifier] = cls
        for iface in self.namespace.get_interfaces():
            for m in iface.methods:
                symbols[m.identifier] = iface
            for m in iface.functions:
                symbols[m.identifier] = iface
        for record in self.namespace.get_records():
            for m in record.constructors:
                symbols[m.identifier] = record
            for m in record.methods:
                symbols[m.identifier] = record
            for m in record.functions:
                symbols[m.identifier] = record
        for union in self.namespace.get_unions():
            for m in union.constructors:
                symbols[m.identifier] = union
            for m in union.methods:
                symbols[m.identifier] = union
            for m in union.functions:
                symbols[m.identifier] = union
        self.namespace._symbols = symbols

    def resolve_interface_implementations(self) -> None:
        seen_impls = {}
        for iface in self.namespace.get_interfaces():
            for cls in self.namespace.get_classes():
                if cls.implements is None:
                    continue
                if iface in cls.implements:
                    seen_impls.setdefault(iface.name, []).append(cls)
        for iface, seen in seen_impls.items():
            if iface in self.namespace._interfaces:
                self.namespace._interfaces[iface].implementations = seen

    def get_class_hierarchy(self, root=None):
        flat_tree = []
        seen_types = {}

        def window(iterable, size=2):
            i = iter(iterable)
            win = []
            for e in range(0, size):
                win.append(next(i))
            yield win
            for e in i:
                win = win[1:] + [e]
                yield win

        for cls in self.namespace.get_classes():
            if cls.parent is None:
                flat_tree.append((cls.name, None))
                continue

            if len(cls.ancestors) < 2:
                flat_tree.append((cls.name, cls.ancestors[0].name))
            else:
                flat_tree.append((cls.name, cls.ancestors[0].name))
                for chunk in window(cls.ancestors, size=2):
                    if chunk[0].name in seen_types:
                        continue
                    if len(chunk) == 2:
                        flat_tree.append((chunk[0].name, chunk[1].name))
                    else:
                        flat_tree.append((chunk[0].name, None))
                    seen_types[chunk[0].name] = 1

        def subtree(cls, rel):
            return {
                v: subtree(v, rel)
                for v in [x[0] for x in rel if x[1] == cls]
            }

        return subtree(root, flat_tree)

    @property
    def namespace(self) -> T.Optional[Namespace]:
        return self._namespaces[0]

    def find_type(self, name: str, ns: T.Optional[str] = None) -> T.Optional[T.Tuple[Namespace, Type]]:
        if ns is None or self.namespace.name == ns:
            res = self.namespace.find_real_type(name)
            if res is not None:
                return (self.namespace, res)
        for repo in self.includes.values():
            if ns is not None and ns != repo.namespace.name:
                continue
            res = repo.namespace.find_real_type(name)
            if res is not None:
                return (repo.namespace, res)
        return None

    def find_symbol(self, name: str) -> T.Optional[T.Tuple[Namespace, Type]]:
        log.debug(f"Looking for symbol {name} in current namespace {self.namespace.name}")
        res = self.namespace.find_symbol(name)
        if res is not None:
            return (self.namespace, res)
        for repo in self.includes.values():
            log.debug(f"Looking for symbol {name} in namespace {repo.namespace.name}")
            res = repo.namespace.find_symbol(name)
            if res is not None:
                return (repo.namespace, res)
        return None

    def find_class(self, name: str, ns: T.Optional[str] = None) -> T.Optional[T.Tuple[Namespace, Type]]:
        if ns is None or self.namespace.name == ns:
            res = self.namespace.find_class(name)
            if res is not None:
                return (self.namespace, res)
        for repo in self.includes.values():
            if ns is not None and ns != repo.namespace.name:
                continue
            res = repo.namespace.find_class(name)
            if res is not None:
                return (repo.namespace, res)
        return None

    def find_interface(self, name: str, ns: T.Optional[str] = None) -> T.Optional[T.Tuple[Namespace, Type]]:
        if ns is None or self.namespace.name == ns:
            res = self.namespace.find_interface(name)
            if res is not None:
                return (self.namespace, res)
        for repo in self.includes.values():
            if ns is not None and ns != repo.namespace.name:
                continue
            res = repo.namespace.find_interface(name)
            if res is not None:
                return (repo.namespace, res)
        return None
